"""A file-based pydap server running on Gunicorn.

Usage:
  pydap [options]

Options:
  -h --help                     Show this help message and exit
  --version                     Show pydap version
  -i --init DIR                 Create directory with templates
  -b ADDRESS --bind ADDRESS     The ip to listen to [default: 127.0.0.1]
  -p PORT --port PORT           The port to connect [default: 8001]
  -d DIR --data DIR             The directory with files [default: .]
  -t DIR --templates DIR        The directory with templates
  --workers INT                 Number of workers [default: 1]
  --threads INT                 Number of threads [default: 1]
  --worker-class=CLASS          Gunicorn worker class [default: sync]
"""

import importlib.resources
import mimetypes
import os
import re
import shutil
from datetime import datetime

from docopt import docopt
from gunicorn.app.wsgiapp import WSGIApplication
from jinja2 import ChoiceLoader, Environment, FileSystemLoader, PackageLoader

# from requests.utils import unquote
from webob import Response
from webob.dec import wsgify
from webob.exc import HTTPForbidden, HTTPNotFound
from webob.static import DirectoryApp, FileApp

from ..exceptions import ExtensionNotSupportedError
from ..handlers.lib import get_handler, load_handlers
from ..lib import __version__, unquote
from .ssf import ServerSideFunctions


class DapServer(object):
    """A directory app that creates file listings and handle DAP requests."""

    def __init__(self, path, templates=None):
        self.path = os.path.abspath(path)

        # the default loader reads templates from the package
        loaders = [PackageLoader("pydap.wsgi", "templates")]

        # optionally, the user can also specify a template directory that will
        # override the default templates; this should have precedence over the
        # default templates
        if templates is not None:
            loaders.insert(0, FileSystemLoader(templates))

        # set the rendering environment; this is also used by pydap responses
        # that need to render templates (like HTML, WMS, KML, etc.)
        self.env = Environment(loader=ChoiceLoader(loaders))
        self.env.filters["datetimeformat"] = datetimeformat
        self.env.filters["datetimeformat_iso"] = datetimeformat_iso
        self.env.filters["unquote"] = unquote

        # cache available handlers, so we don't need to load them every request
        self.handlers = load_handlers()

    @wsgify
    def __call__(self, req):
        """WSGI application callable.

        Returns either a file download, directory listing or DAP response.

        """
        path = os.path.abspath(os.path.join(self.path, *req.path_info.split("/")))

        if not path.startswith(self.path):
            return HTTPForbidden()
        if path.endswith("catalog.xml"):
            return self.index(os.path.dirname(path), req, catalog=True)
        elif os.path.exists(path):
            if os.path.isdir(path):
                return self.index(path, req)
            else:
                return FileApp(path)

        # strip DAP extension (``.das``, eg) and see if the file exists
        base, ext = os.path.splitext(path)
        if os.path.isfile(base):
            req.environ["pydap.jinja2.environment"] = self.env
            app = ServerSideFunctions(get_handler(base, self.handlers))
            return req.get_response(app)
        else:
            return HTTPNotFound(comment=path)

    def index(self, directory, req, catalog=False):
        """Return a directory listing."""

        template_name = "index.html"
        response_content_type = "text/html"
        content = [os.path.join(directory, name) for name in os.listdir(directory)]

        files = [
            {
                "name": os.path.split(path)[1],
                "size": os.path.getsize(path),
                "last_modified": datetime.fromtimestamp(os.path.getmtime(path)),
                "supported": supported(path, self.handlers),
            }
            for path in content
            if os.path.isfile(path)
        ]
        files.sort(key=lambda d: alphanum_key(d["name"]))

        directories = [
            {
                "name": os.path.split(path)[1],
                "last_modified": datetime.fromtimestamp(os.path.getmtime(path)),
            }
            for path in content
            if os.path.isdir(path)
        ]
        directories.sort(key=lambda d: alphanum_key(d["name"]))

        tokens = req.path_info.split("/")[1:]
        breadcrumbs = [
            {
                "url": "/".join([req.application_url] + tokens[: i + 1]),
                "title": token,
            }
            for i, token in enumerate(tokens)
            if token
        ]

        context = {
            "root": req.application_url,
            "location": req.path_url,
            "breadcrumbs": breadcrumbs,
            "directories": directories,
            "files": files,
            "version": __version__,
        }

        if catalog:
            context["location"] = req.path.replace("catalog.xml", "")
            template_name = "catalog.xml"
            response_content_type = "application/xml"

        template = self.env.get_template(template_name)
        return Response(
            body=template.render(context),
            content_type=response_content_type,
            charset="utf-8",
        )


def supported(filepath, handlers=None):
    """Test if a file has a corresponding handler.

    Returns a boolean.

    """
    try:
        get_handler(filepath, handlers, instantiate=False)
        return True
    except ExtensionNotSupportedError:
        return False


def alphanum_key(s):
    """Parse a string, returning a list of string and number chunks.

        >>> alphanum_key("z23a")
        ['z', 23, 'a']

    Useful for sorting names in a natural way.

    From http://nedbatchelder.com/blog/200712.html#e20071211T054956

    """

    def tryint(s):
        try:
            return int(s)
        except Exception:
            return s

    return [tryint(c) for c in re.split("([0-9]+)", s)]


def datetimeformat(value, format="%Y-%m-%d %H:%M:%S"):
    """Return a formatted datetime object."""
    return value.strftime(format)


def datetimeformat_iso(value):
    """Return a formatted datetime object as ISO8601 representation."""
    return value.isoformat()


class StaticMiddleware(object):
    """WSGI middleware for static assets.

    The assets can be either specified as a directory, or retrieved from a
    Python package. Inspired by ``werkezeug.wsgi.SharedDataMiddleware``.

    """

    def __init__(self, app, static):
        self.app = app
        self.static = static

    @wsgify
    def __call__(self, req):
        if req.path_info_peek() != "static":
            return req.get_response(self.app)

        # strip "/static"
        req.path_info_pop()

        # statically serve the directory
        if isinstance(self.static, str):
            return req.get_response(DirectoryApp(self.static))

        # otherwise, load resource from package
        package, resource_path = self.static
        parts = resource_path.split("/")
        resources_path = importlib.resources.files(package)
        for i in range(len(parts)):
            resources_path /= parts[i]
        resource = req.path_info.split("/")[-1]
        full_path = resources_path / resource

        if not full_path.is_file():
            return HTTPNotFound(req.path_info)

        content_type, content_encoding = mimetypes.guess_type(full_path)
        package += (".").join([""] + parts)
        return Response(
            body=importlib.resources.read_text(package, resource),
            content_type=content_type,
            content_encoding=content_encoding,
        )


def init(directory):
    """Create directory with default templates."""
    # copy main templates
    templates = importlib.resources.files("pydap.wsgi") / "templates"
    with importlib.resources.as_file(templates) as path:
        shutil.copytree(templates, directory)

    # copy templates from HTML response
    for resource in importlib.resources.files(
        "pydap.responses.html.templates"
    ).iterdir():
        path = importlib.resources.files(
            "pydap.responses.html"
        ) / "templates/{0}".format(resource.name)
        with importlib.resources.as_file(path) as Path:
            shutil.copy(Path, directory)


class PyDapApplication(WSGIApplication):
    """An application interface for configuring and loading
    the various necessities for a given web framework."""

    def __init__(self, app, **local_config):

        self._app = app
        self._config = local_config
        self._config["bind"] = "%s:%s" % (
            local_config.pop("host", ""),
            local_config.pop("port", ""),
        )

        WSGIApplication.__init__(self)

    def load_config(self):
        for k, v in self._config.items():
            if v is not None:
                self.cfg.set(k.lower(), v)

    def load(self):
        return self._app


def main():  # pragma: no cover
    """Run server from the command line."""

    arguments = docopt(__doc__, version="pydap %s" % __version__)

    # init templates?
    if arguments["--init"]:
        init(arguments["--init"])
        return

    # create pydap app
    data, templates = arguments["--data"], arguments["--templates"]
    app = DapServer(data, templates)

    # configure app so that is reads static assets from the template directory
    # or from the package
    if templates and os.path.exists(os.path.join(templates, "static")):
        static = os.path.join(templates, "static")
    else:
        static = ("pydap.wsgi", "templates/static")
    app = StaticMiddleware(app, static)

    # configure WSGI server
    pydap_app = PyDapApplication(
        app,
        host=arguments["--bind"],
        port=int(arguments["--port"]),
        workers=arguments["--workers"],
        threads=arguments["--threads"],
        worker_class=arguments["--worker-class"],
    )
    pydap_app.run()


if __name__ == "__main__":
    main()



"""The DODS response.

This is the DAP response that carries data. The response comes with a DDS
header describing the structure of the data, followed by the data encoded as
XDR.

Even though Python has a library for XDR encoding/decoding, the DODS response
uses Numpy directly since it's faster.

"""

import copy

import numpy as np

from ..lib import (
    DAP2_ARRAY_LENGTH_NUMPY_TYPE,
    DAP2_TO_NUMPY_RESPONSE_TYPEMAP,
    END_OF_SEQUENCE,
    NUMPY_TO_DAP2_TYPEMAP,
    START_OF_SEQUENCE,
    __version__,
    walk,
)
from ..model import BaseType, SequenceType, StructureType
from .dds import dds
from .lib import BaseResponse

try:
    from functools import singledispatch
except ImportError:
    from singledispatch import singledispatch


def DAP2_response_dtypemap(dtype):
    """
    This function takes a numpy dtype object
    and returns a dtype object that is compatible with
    the DAP2 specification.
    """
    dtype_str = DAP2_TO_NUMPY_RESPONSE_TYPEMAP[NUMPY_TO_DAP2_TYPEMAP[dtype.char]]
    return np.dtype(dtype_str)


def tostring_with_byteorder(x, dtype):
    order = dtype.byteorder
    _var = x.astype(dtype.str)
    return _var.view(_var.dtype.newbyteorder(order)).tobytes()


class DODSResponse(BaseResponse):
    """The DODS response."""

    __version__ = __version__

    def __init__(self, dataset):
        BaseResponse.__init__(self, dataset)
        self.headers.extend(
            [
                ("Content-description", "dods_data"),
                ("Content-type", "application/octet-stream"),
            ]
        )

        length = calculate_size(dataset)
        if length is not None:
            self.headers.append(("Content-length", str(length)))

    def __iter__(self):
        # generate DDS
        for line in dds(self.dataset):
            yield line.encode("ascii")

        yield b"Data:\n"
        for block in dods(self.dataset):
            yield block


@singledispatch
def dods(var):
    """Single dispatcher for generating the DODS response."""
    raise StopIteration


@dods.register(StructureType)
def _structuretype(var):
    for child in var.children():
        for block in dods(child):
            yield block


@dods.register(SequenceType)
def _sequencetype(var):
    # a flat array can be processed one record (or more?) at a time
    if all(isinstance(child, BaseType) for child in var.children()):
        DAP2_types = []
        position = 0
        try:
            for child in var.children():
                if DAP2_response_dtypemap(child.dtype).char == "S":
                    (DAP2_types.append(DAP2_ARRAY_LENGTH_NUMPY_TYPE))  # string length
                    DAP2_types.append("|S{%s}" % position)  # string padded to 4n
                    position += 1
                else:
                    # Convert any numpy dtypes to numpy dtypes compatible
                    # with the DAP2 standard:
                    DAP2_types.append(DAP2_response_dtypemap(child.dtype).str)
        except StopIteration:
            return
        DAP2_dtype_str = ",".join(DAP2_types)
        strings = position > 0

        # array initializations is costy, so we keep a cache here; this will
        # be inneficient if there are many strings of different length only
        cache = {}

        for record in var.iterdata():
            yield START_OF_SEQUENCE

            if strings:
                out = []
                padded = []
                for value in record:
                    if isinstance(value, str):
                        length = len(value) or 1
                        out.append(length)
                        padded.append(length + (-length % 4))
                    out.append(value)
                record = out
                DAP2_dtype_str = ",".join(DAP2_types).format(*padded)

            if DAP2_dtype_str not in cache:
                # Remember that DAP2_dtype is a (possibly composite)
                # numpy dtype that is compatible with the DAP2
                # data model. This means that all dtypes in
                # DAP2_dtype are representable in DAP2 -- AND --
                # the data in var can all be upconverted
                # in a lossless manner to the dtypes in DAP2_dtype.
                cache[DAP2_dtype_str] = np.zeros((1,), dtype=DAP2_dtype_str)
            # By assigning record to ``cache`` the upconversion
            # occurs naturally:
            cache[DAP2_dtype_str][:] = tuple(record)
            # byteorder was taken care of during the upconversion:
            yield cache[DAP2_dtype_str].tobytes()

        yield END_OF_SEQUENCE

    # nested array, need to process individually
    else:
        # create a template structure
        struct = StructureType(var.name)
        for name in var.keys():
            struct[name] = copy.copy(var[name])

        for record in var.iterdata():
            yield START_OF_SEQUENCE
            struct.data = record
            for block in dods(struct):
                yield block
        yield END_OF_SEQUENCE


@dods.register(BaseType)
def _basetype(var):
    data = var.data

    if not hasattr(data, "shape"):
        data = np.asarray(data)

    # Convert any numpy dtypes to numpy dtypes compatible
    # with the DAP2 standard:
    DAP2_dtype = DAP2_response_dtypemap(data.dtype)

    if data.shape:
        # pack length for arrays
        length = np.prod(data.shape).astype(int)

        # send length twice at the begining of an array...
        factor = 2
        if DAP2_dtype.char == "S":
            # ... expcept for strings:
            factor = 1
        yield tostring_with_byteorder(
            length, np.dtype(DAP2_ARRAY_LENGTH_NUMPY_TYPE)
        ) * factor

    # make data iterable; 1D arrays must be converted to 2D, since
    # iteration over 1D yields scalars which are not properly cast to big
    # endian
    # This line was removed because endianness is now treated explicitly
    # in tostring_with_byteorder()
    # if len(data.shape) < 2:
    #     data = data.reshape(1, -1)
    # Only ensure that 0d arrays are iterable:
    if len(data.shape) == 0:
        data = data[np.newaxis]

    # unsigned bytes are padded up to 4n
    if DAP2_dtype == np.ubyte:
        length = np.prod(data.shape).astype(int)
        for block in data:
            yield tostring_with_byteorder(block, DAP2_dtype)
        yield (-length % 4) * b"\0"

    # regular data
    else:
        # strings are also zero padded and preceeded by their length
        if DAP2_dtype.char == "S":
            for block in data:
                for word in block.flat:
                    length = len(word)
                    yield tostring_with_byteorder(
                        np.array(length), np.dtype(DAP2_ARRAY_LENGTH_NUMPY_TYPE)
                    )
                    # byteorder is not important for strings:
                    if hasattr(word, "encode"):
                        yield word.encode("ascii")
                    elif hasattr(word, "tobytes"):
                        yield word.tobytes()
                    else:
                        raise TypeError(
                            "Could not convert word '{0}' to bytes".format(word)
                        )
                    yield (-length % 4) * b"\0"
        else:
            for block in data:
                # Remember that DAP2_dtype is a
                # numpy dtype that is compatible with the DAP2
                # data model. This means that the dtype in
                # DAP2_dtype is representable in DAP2 -- AND --
                # the data in var can all be upconverted
                # in a lossless manner to the dtype in DAP2_dtype.
                yield tostring_with_byteorder(block, DAP2_dtype)


def calculate_size(dataset):
    """Calculate the size of the response. Returns the size in bytes."""
    length = 0

    for var in walk(dataset):
        # pydap can't calculate the size of sequences since the data is
        # streamed directly from the source. Also, strings are encoded
        # individually, so it's not possible to get their size unless we read
        # everything.
        if isinstance(var, SequenceType) or (
            isinstance(var, BaseType) and DAP2_response_dtypemap(var.dtype).char == "S"
        ):
            return None
        elif isinstance(var, BaseType):
            if var.shape:
                length += 8  # account for array size marker

            size = int(np.prod(var.shape))

            # Convert any numpy dtype to numpy dtype compatible
            # with the DAP2 standard:
            DAP2_dtype = DAP2_response_dtypemap(var.data.dtype)
            if DAP2_dtype == np.ubyte:
                length += size + (-size % 4)
            else:
                # Remember that numpy dtypes are upconverted to
                # DAP2_dtype when sent in the response.
                # Their length must thus be modified accordingly:
                length += size * DAP2_dtype.itemsize

    # account for DDS
    length += len("".join(dds(dataset))) + len(b"Data:\n")
    return length



"""The DMR response.

The DMR response builds a representation of the structure of the dataset,
informing which variables are contained, their shape, type and dimensions.
The DMR describes all metadata associated with a given dataset, allowing
clients to introspect the variables and request data as necessary.

"""

try:
    from functools import singledispatch
except ImportError:
    from singledispatch import singledispatch

from ..lib import __version__
from ..model import BaseType, DatasetType, GroupType, SequenceType, StructureType
from .lib import BaseResponse

INDENT = " " * 4
namespace = {"": "http://xml.opendap.org/ns/DAP/4.0#"}


class DMRResponse(BaseResponse):
    """The DDS response."""

    __version__ = __version__

    def __init__(self, dataset):
        BaseResponse.__init__(self, dataset)
        self.headers.extend(
            [
                ("Content-description", "dmr"),
                ("Content-type", "text/plain; charset=ascii"),
            ]
        )

    def __iter__(self):
        # generate DMR
        yield '<?xml version="1.0" encoding="ISO-8859-1"?>\n'.encode("ascii")
        for line in dmr(self.dataset):
            yield line.encode("ascii")


@singledispatch
def dmr(var):
    """Single dispatcher for generating the DODS response."""
    raise StopIteration


@dmr.register(DatasetType)
def _(var, level=0):
    str0 = 'Dataset xmlns="{namespace}"'.format(namespace=namespace[""])
    str1 = ' xml:base="{url}"'.format(url="http://localhost:8001")
    str2 = ' dapVersion="4.0" dmrVersion="1.0"'
    str3 = ' name="{name}">\n'.format(name=var.name)
    yield "<{indent}".format(indent=level * INDENT) + str0 + str1 + str2 + str3
    for name, size in var.dimensions.items():
        yield '{indent}<Dimension name="{name}" size="{size}"/>\n'.format(
            indent=(level + 1) * INDENT, name=name, size=size
        )
    for child in var.children():
        for line in dmr(child, level + 1):
            yield line
    for key, value in var.attributes.items():
        if key != "dimensions":
            _type = type(value).__name__
            yield '{indent}<Attribute name="{name}" type="{type}">\n'.format(
                indent=(level + 1) * INDENT, name=key, type=_type
            )
            yield "{indent}<Value>{val}</Value>\n".format(
                indent=(level + 2) * INDENT, val=value
            )
            yield "{indent}</Attribute>\n".format(indent=(level + 1) * INDENT)

    yield "{indent}</Dataset>\n".format(indent=level * INDENT)


@dmr.register(StructureType)
def _structuretype(var):
    pass


@dmr.register(SequenceType)
def _sequencetype(var, level=0, sequence=0):
    pass


@dmr.register(GroupType)
def _grouptype(var, level=0):
    yield '{indent}<Group name="{name}">\n'.format(indent=level * INDENT, name=var.name)
    for dim, size in var.attributes["dimensions"].items():
        yield '{indent}<Dimension name="{name}" size="{size}"/>\n'.format(
            indent=(level + 1) * INDENT, name=dim, size=size
        )
    for key, value in var.attributes.items():
        if key not in ["dimensions", "path"]:
            _type = type(value).__name__
            yield '{indent}<Attribute name="{name}" type="{type}">\n'.format(
                indent=(level + 1) * INDENT, name=key, type=_type
            )
            yield "{indent}<Value>{val}</Value>\n".format(
                indent=(level + 2) * INDENT, val=value
            )
            yield "{indent}</Attribute>\n".format(indent=(level + 1) * INDENT)
    for child in var.children():
        for line in dmr(child, level + 1):
            yield line
    yield "{indent}</Group>\n".format(indent=level * INDENT)


@dmr.register(BaseType)
def _basetype(var, level=0):
    _ntype = var.dtype
    _vartype = str(_ntype)[0].upper() + str(_ntype)[1:]
    if _vartype == "<U0":
        _vartype = "String"
    yield '{indent}<{type} name="{name}">\n'.format(
        indent=level * INDENT,
        type=_vartype,
        name=var.name,
    )
    # get dimensions
    for dim in var.dimensions:
        yield '{indent}<Dim name="{name}"/>\n'.format(
            indent=(level + 1) * INDENT, name=dim
        )
    for key, value in var.attributes.items():
        if key != "dims":
            _type = type(value).__name__
            yield '{indent}<Attribute name="{name}" type="{type}">\n'.format(
                indent=(level + 1) * INDENT, name=key, type=_type
            )
            yield "{indent}<Value>{val}</Value>\n".format(
                indent=(level + 2) * INDENT, val=value
            )
            yield "{indent}</Attribute>\n".format(indent=(level + 1) * INDENT)
    yield "{indent}</{type}>\n".format(indent=level * INDENT, type=_vartype)

    # get maps
    if "Maps" in var.attributes.keys():
        Maps = var.Maps
        for _map in Maps:
            yield '{indent}<Map name="{name}">\n'.format(
                indent=(level + 1) * INDENT, name=_map
            )



"""A DDS parser."""

import re

import numpy as np

from ..lib import LOWER_DAP2_TO_NUMPY_PARSER_TYPEMAP, _quote
from ..model import BaseType, DatasetType, GridType, SequenceType, StructureType
from . import SimpleParser

constructors = ("grid", "sequence", "structure")
name_regexp = r'[\w%!~"\'\*-]+'


def DAP2_parser_typemap(type_string):
    """
    This function takes a numpy dtype object
    and returns a dtype object that is compatible with
    the DAP2 specification.
    """
    dtype_str = LOWER_DAP2_TO_NUMPY_PARSER_TYPEMAP[type_string.lower()]
    return np.dtype(dtype_str)


class DDSParser(SimpleParser):
    """A parser for the DDS."""

    def __init__(self, dds):
        super(DDSParser, self).__init__(dds, re.IGNORECASE)
        self.dds = dds

    def consume(self, regexp):
        """Consume and return a token."""
        token = super(DDSParser, self).consume(regexp)
        self.buffer = self.buffer.lstrip()
        return token

    def parse(self):
        """Parse the DAS, returning a dataset."""
        dataset = DatasetType("nameless")

        self.consume("dataset")
        self.consume("{")
        while not self.peek("}"):
            var = self.declaration()
            dataset[var.name] = var
        self.consume("}")

        dataset.name = _quote(self.consume("[^;]+"))
        dataset._set_id(dataset.name)
        self.consume(";")

        return dataset

    def declaration(self):
        """Parse and return a declaration."""
        token = self.peek(r"\w+").lower()

        map = {
            "grid": self.grid,
            "sequence": self.sequence,
            "structure": self.structure,
        }
        method = map.get(token, self.base)
        return method()

    def base(self):
        """Parse a base variable, returning a ``BaseType``."""
        data_type_string = self.consume(r"\w+")

        parser_dtype = DAP2_parser_typemap(data_type_string)
        name = _quote(self.consume(r"[^;\[]+"))

        shape, dimensions = self.dimensions()
        self.consume(r";")

        data = DummyData(parser_dtype, shape)
        var = BaseType(name, data, dimensions=dimensions)

        return var

    def dimensions(self):
        """Parse variable dimensions, returning tuples of dimensions/names."""
        shape = []
        names = []
        while not self.peek(";"):
            self.consume(r"\[")
            token = self.consume(name_regexp)
            if self.peek("="):
                names.append(token)
                self.consume("=")
                token = self.consume(r"\d+")
            shape.append(int(token))
            self.consume(r"\]")
        return tuple(shape), tuple(names)

    def sequence(self):
        """Parse a DAS sequence, returning a ``SequenceType``."""
        sequence = SequenceType("nameless")
        self.consume("sequence")
        self.consume("{")

        while not self.peek("}"):
            var = self.declaration()
            sequence[var.name] = var
        self.consume("}")

        sequence.name = _quote(self.consume("[^;]+"))
        self.consume(";")
        return sequence

    def structure(self):
        """Parse a DAP structure, returning a ``StructureType``."""
        structure = StructureType("nameless")
        self.consume("structure")
        self.consume("{")

        while not self.peek("}"):
            var = self.declaration()
            structure[var.name] = var
        self.consume("}")

        structure.name = _quote(self.consume("[^;]+"))
        self.consume(";")

        return structure

    def grid(self):
        """Parse a DAP grid, returning a ``GridType``."""
        grid = GridType("nameless")
        self.consume("grid")
        self.consume("{")

        self.consume("array")
        self.consume(":")
        array = self.base()
        grid[array.name] = array

        self.consume("maps")
        self.consume(":")
        while not self.peek("}"):
            var = self.base()
            grid[var.name] = var
        self.consume("}")

        grid.name = _quote(self.consume("[^;]+"))
        self.consume(";")

        return grid


def dds_to_dataset(dds):
    """Return a dataset object from a DDS representation."""
    return DDSParser(dds).parse()


class DummyData(object):
    def __init__(self, dtype, shape):
        self.dtype = dtype
        self.shape = shape



"""Test IterData, which makes iterables behave like Numpy structured arrays."""

import numpy as np
import pytest

from pydap.handlers.lib import IterData
from pydap.model import BaseType, DatasetType, SequenceType


@pytest.fixture
def simple_array():
    arr = np.array(
        [
            (0, 1, 10.0),
            (1, 2, 20.0),
            (2, 3, 30.0),
            (3, 4, 40.0),
            (4, 5, 50.0),
            (5, 6, 60.0),
            (6, 7, 70.0),
            (7, 8, 80.0),
        ],
        dtype=[("byte", "b"), ("int", "i4"), ("float", "f4")],
    )
    return arr


@pytest.fixture
def simple_object(simple_array):
    # add sequence and children for each column
    name = "nameless"
    dataset = DatasetType(name)
    seq = dataset["sequence"] = SequenceType("sequence")
    for var in simple_array.dtype.names:
        seq[var] = BaseType(var)

    obj = IterData(
        [
            (0, 1, 10.0),
            (1, 2, 20.0),
            (2, 3, 30.0),
            (3, 4, 40.0),
            (4, 5, 50.0),
            (5, 6, 60.0),
            (6, 7, 70.0),
            (7, 8, 80.0),
        ],
        seq,
    )
    return obj


def test_iter(simple_array, simple_object):
    assert [tuple(row) for row in simple_array] == [
        (0, 1, 10.0),
        (1, 2, 20.0),
        (2, 3, 30.0),
        (3, 4, 40.0),
        (4, 5, 50.0),
        (5, 6, 60.0),
        (6, 7, 70.0),
        (7, 8, 80.0),
    ]
    assert list(simple_array["byte"]) == [0, 1, 2, 3, 4, 5, 6, 7]

    assert [tuple(row) for row in simple_object] == [
        (0, 1, 10.0),
        (1, 2, 20.0),
        (2, 3, 30.0),
        (3, 4, 40.0),
        (4, 5, 50.0),
        (5, 6, 60.0),
        (6, 7, 70.0),
        (7, 8, 80.0),
    ]
    assert list(simple_object["byte"]) == [0, 1, 2, 3, 4, 5, 6, 7]


def test_dtype(simple_array, simple_object):
    assert simple_array.dtype == np.dtype(
        [("byte", "i1"), ("int", "<i4"), ("float", "<f4")]
    )
    assert simple_array["int"].dtype == "<i4"

    assert simple_object.dtype == np.dtype(
        [("byte", "<i8"), ("int", "<i8"), ("float", "<f8")]
    )
    assert simple_object["int"].dtype == "<i8"


def test_selection(simple_array, simple_object):
    selection = simple_array[simple_array["byte"] > 3]
    assert [tuple(row) for row in selection] == [
        (4, 5, 50.0),
        (5, 6, 60.0),
        (6, 7, 70.0),
        (7, 8, 80.0),
    ]

    selection = simple_object[simple_object["byte"] > 3]
    assert [tuple(row) for row in selection] == [
        (4, 5, 50.0),
        (5, 6, 60.0),
        (6, 7, 70.0),
        (7, 8, 80.0),
    ]


def test_projection(simple_array, simple_object):
    projection = simple_array[1::2]
    assert [tuple(row) for row in projection] == [
        (1, 2, 20.0),
        (3, 4, 40.0),
        (5, 6, 60.0),
        (7, 8, 80.0),
    ]
    assert list(simple_array["byte"][1::2]) == [1, 3, 5, 7]
    assert list(simple_array[1::2]["byte"]) == [1, 3, 5, 7]

    projection = simple_object[1::2]
    assert [tuple(row) for row in projection] == [
        (1, 2, 20.0),
        (3, 4, 40.0),
        (5, 6, 60.0),
        (7, 8, 80.0),
    ]
    assert list(simple_object["byte"][1::2]) == [1, 3, 5, 7]
    assert list(simple_object[1::2]["byte"]) == [1, 3, 5, 7]


def test_combined(simple_array, simple_object):
    filtered = simple_array[simple_array["byte"] > 1]
    filtered = filtered[filtered["byte"] < 6]
    filtered = filtered[::2]
    assert [tuple(row) for row in filtered] == [(2, 3, 30.0), (4, 5, 50.0)]

    filtered = simple_object[simple_object["byte"] > 1]
    filtered = filtered[filtered["byte"] < 6]
    filtered = filtered[::2]
    assert [tuple(row) for row in filtered] == [(2, 3, 30.0), (4, 5, 50.0)]


def test_combined_other(simple_array, simple_object):
    filtered = simple_array[["int", "float"]][simple_array["byte"] < 2]
    assert [tuple(row) for row in filtered] == [(1, 10.0), (2, 20.0)]

    filtered = simple_object[["int", "float"]][simple_object["byte"] < 2]
    assert [tuple(row) for row in filtered] == [(1, 10.0), (2, 20.0)]


@pytest.fixture
def nested_data():
    shallow_data = [(1, 1, 1), (2, 4, 4), (3, 6, 9), (4, 8, 16)]
    deep_data = [
        [(10, 11, 12), (21, 22, 23)],
        [(15, 16, 17)],
        [],
        [(31, 32, 33), (41, 42, 43), (51, 52, 53), (61, 62, 63)],
    ]
    nested = [x + (deep_data[x_id],) for x_id, x in enumerate(shallow_data)]
    return nested


@pytest.fixture
def nested_dtype():
    dtype = np.dtype(
        [
            ("a", "<i8"),
            ("b", "<i8"),
            ("c", "<i8"),
            ("d", np.dtype([("e", "<i8"), ("f", "<i8"), ("g", "<i8")])),
        ]
    )
    return dtype


@pytest.fixture
def nested_object(nested_data):
    name = "nameless"
    dataset = DatasetType(name)
    seq = dataset["nested"] = SequenceType("nested")
    for var in ["a", "b", "c"]:
        seq[var] = BaseType(var)
    seq["d"] = SequenceType("d")
    for var in ["e", "f", "g"]:
        seq["d"][var] = BaseType(var)

    nested = IterData(nested_data, seq)
    return nested


def test_nested_iter(nested_data, nested_object):
    assert [tuple(row) for row in nested_object] == nested_data


def test_nested_iter_child(nested_data, nested_object):
    expected = [row[0] for row in nested_data]
    assert list(nested_object["a"]) == expected


def test_nested_iter_nested_sequence(nested_data, nested_object):
    expected = [row[3] for row in nested_data]
    assert list(nested_object["d"]) == expected


def test_nested_iter_nested_deep_child(nested_data, nested_object):
    expected = [[col[0] for col in row[3]] for row in nested_data]
    assert list(nested_object["d"]["e"]) == expected


def test_nested_dtype(nested_dtype, nested_object):
    assert nested_object.dtype == nested_dtype


def test_nested_selection(nested_data, nested_object):
    selection = nested_object[nested_object["a"] > 2]
    expected = [tuple(row) for row in nested_data if row[0] > 2]
    assert [tuple(row) for row in selection] == expected


def test_nested_projection(nested_data, nested_object):
    projection = nested_data[1::2]
    expected = [
        tuple(row)
        for row_id, row in enumerate(nested_data)
        if row_id in range(1, len(nested_data), 2)
    ]
    assert [tuple(row) for row in projection] == expected

    projection = nested_object[1::2]
    expected = [
        tuple(row)
        for row_id, row in enumerate(nested_data)
        if row_id in range(1, len(nested_data), 2)
    ]
    assert [tuple(row) for row in projection] == expected


def test_nested_combined(nested_data, nested_object):
    filtered = [tuple(row) for row in nested_data if row[0] > 2]
    filtered = [tuple(row) for row in filtered if row[0] < 4]
    filtered = filtered[::2]
    assert [tuple(row) for row in filtered] == [(3, 6, 9, [])]

    filtered = nested_object[nested_object["a"] > 2]
    filtered = filtered[filtered["a"] < 4]
    filtered = filtered[::2]
    assert [tuple(row) for row in filtered] == [(3, 6, 9, [])]



"""Basic functions for handlers.

pydap handlers are responsible for reading data in different formats -- NetCDF,
SQL databases, CSV files, etc. -- and convert them into the internal data model
so that the data may be served using different responses.

"""

from __future__ import division

import ast
import copy
import itertools
import operator
import re
import sys
from importlib.metadata import entry_points

import numpy as np
from numpy.lib import Arrayterator
from webob import Request

from pydap.exceptions import ConstraintExpressionError, ExtensionNotSupportedError
from pydap.lib import encode, fix_shorthand, get_var, walk
from pydap.model import BaseType, DatasetType, GridType, SequenceType, StructureType
from pydap.parsers import parse_ce, parse_selection
from pydap.responses.error import ErrorResponse
from pydap.responses.lib import load_responses

# buffer size in bytes, for streaming data
BUFFER_SIZE = 2**27

CORS_RESPONSES = ["dds", "das", "dods", "ver", "json", "dmr"]


def load_handlers():
    r"""Load all handlers, returning them on a list."""
    eps = entry_points(group="pydap.handler")
    Rs = [r for r in eps if r.module[:5] == "pydap"]
    nRs = [r for r in eps if r.module[:5] != "pydap"]
    base_dict = dict((r.name, r.load()) for r in Rs)

    opts_dict = dict((r.name, r.load()) for r in nRs)
    base_dict.update(opts_dict)
    return base_dict.values()


def get_handler(filepath, handlers=None, instantiate=True):
    """Given a filepath, return the corresponding instantiated handler."""
    # Check each handler to see which one handles this file.
    for handler in handlers or load_handlers():
        p = re.compile(handler.extensions)
        if p.match(filepath):
            # only check if extension is supported - don't return instance
            if not instantiate:
                return None
            return handler(filepath)

    raise ExtensionNotSupportedError(
        "No handler available for file {filepath}.".format(filepath=filepath)
    )


class BaseHandler(object):
    """Base class for pydap handlers.

    Handlers are WSGI applications that parse the client request and build the
    corresponding dataset. The dataset is passed to proper Response (DDS, DAS,
    etc.)

    """

    # load all available responses
    #    import pdb; pdb.set_trace()
    responses = load_responses()

    def __init__(self, dataset=None, gzip=False):
        self.dataset = dataset
        self.additional_headers = []
        self._gzip = gzip

    def __call__(self, environ, start_response):
        req = Request(environ)
        path, response = req.path.rsplit(".", 1)
        if response == "das":
            req.query_string = ""
        projection, selection = parse_ce(req.query_string)
        buffer_size = environ.get("pydap.buffer_size", BUFFER_SIZE)

        try:
            # build the dataset and pass it to the proper response, returning a
            # WSGI app
            dataset = self.parse(projection, selection, buffer_size)
            app = self.responses[response](dataset)
            app.close = self.close

            # now build a Response and set additional headers
            res = req.get_response(app)
            for key, value in self.additional_headers:
                res.headers.add(key, value)

            # CORS for Javascript requests
            if response in CORS_RESPONSES:
                res.headers.add("Access-Control-Allow-Origin", "*")
                res.headers.add(
                    "Access-Control-Allow-Headers",
                    "Origin, X-Requested-With, Content-Type",
                )

            if self._gzip:
                res.encode_content()
            return res(environ, start_response)
        except Exception:
            # should the exception be catched?
            if environ.get("x-wsgiorg.throw_errors"):
                raise
            else:
                res = ErrorResponse(info=sys.exc_info())
                return res(environ, start_response)

    def parse(self, projection, selection, buffer_size=BUFFER_SIZE):
        """Parse the constraint expression, returning a new dataset."""
        if self.dataset is None:
            raise NotImplementedError(
                "Subclasses must define a ``dataset`` attribute pointing to a"
                "``DatasetType`` object."
            )

        # make a copy of the dataset, so we can filter sequences inplace
        dataset = copy.copy(self.dataset)

        # apply the selection to the dataset, inplace
        apply_selection(selection, dataset)

        # wrap data in Arrayterator, to optimize projection/selection
        dataset = wrap_arrayterator(dataset, buffer_size)

        # fix projection
        if projection:
            projection = fix_shorthand(projection, dataset)
        else:
            projection = [[(key, ())] for key in dataset.keys()]
        dataset = apply_projection(projection, dataset)

        return dataset

    def close(self):
        """Optional method for closing the dataset."""
        pass


def wrap_arrayterator(dataset, size):
    """Wrap `BaseType` objects in an Arrayterator.

    This function is used to optimize access to huge datasets. It returns a new
    dataset with data wrapped in Arrayterators. This way the data is read in
    blocks instead of buffering everything in memory.

    Since the buffer size of the Arrayterator is in elements, not bytes, we
    convert according to the data item size.

    """
    for var in walk(dataset, BaseType):
        if (
            not isinstance(var.data, Arrayterator)
            and var.data.dtype.itemsize
            and var.data.shape
        ):
            elements = size // var.data.dtype.itemsize
            var.data = Arrayterator(var.data, elements)

    return dataset


def apply_selection(selection, dataset):
    """Apply a given selection to a dataset, modifying it inplace.

    Returns the original dataset.

    """
    for seq in walk(dataset, SequenceType):
        # apply only relevant selections
        conditions = [
            condition
            for condition in selection
            if re.match(r"%s\.[^\.]+(<=|<|>=|>|=|!=)" % re.escape(seq.id), condition)
        ]
        for condition in conditions:
            id1, op, id2 = parse_selection(condition, dataset)
            seq.data = seq[op(id1, id2)].data
    return dataset


def degenerate_grid_to_structure(candidate):
    if isinstance(candidate, GridType):
        candidate = StructureType(candidate.name, candidate.attributes)
    return candidate


def apply_projection(projection, dataset):
    """Apply a given projection to a dataset.

    This function builds and returns a new dataset by adding those variables
    that were requested on the projection.
    """
    out = DatasetType(name=dataset.name, attributes=dataset.attributes)

    # first collect all the variables
    for p in projection:
        target, template = out, dataset
        for i, (name, slice_) in enumerate(p):
            candidate = template[name]

            # add variable to target
            if isinstance(candidate, StructureType):
                if name not in target.keys():
                    if i < len(p) - 1:
                        # A shallow copy of the candidate is created
                        candidate = candidate.__shallowcopy__()
                        # Grids are degenerated into Structures
                        candidate = degenerate_grid_to_structure(candidate)
                    target[name] = candidate
                target, template = target[name], template[name]
            else:
                target[name] = candidate

    # fix sequence data to include only variables that are in the sequence
    for seq in walk(out, SequenceType):
        seq.data = get_var(dataset, seq.id)[tuple(seq.keys())].data

    # apply slices
    for p in projection:
        target = out
        for name, slice_ in p:
            target, parent = target[name], target

            if slice_:
                if isinstance(target, BaseType):
                    target.data = target[slice_]
                elif isinstance(target, SequenceType):
                    parent[name] = target[slice_[0]]
                elif isinstance(target, GridType):
                    parent[name] = target[slice_]
                else:
                    raise ConstraintExpressionError("Invalid projection!")

    return out


class ConstraintExpression(object):
    """An object representing a selection on a constraint expression.

    These can be accumulated so that they are evaluated only once.

    """

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return str(self.value)

    def __and__(self, other):
        """Join two CEs together, returning a new object."""
        return self.__class__(self.value + "&" + str(other))

    def __or__(self, other):
        raise ConstraintExpressionError(
            "OR constraints not allowed in the Opendap specification."
        )


class IterData(object):
    """Class for manipulating data streams as structured arrays.

    A structured array is a Numpy construct that has some very interesting
    properties for working with tabular data.

    """

    shape = ()

    def __init__(self, stream, template, ifilter=None, imap=None, islice=None, level=0):
        self.stream = stream
        self.template = template
        self.level = level

        # these are used to lazily evaluate the data stream
        self.ifilter = ifilter or []
        self.imap = imap or [fix_nested(template)]
        self.islice = islice or []

    @property
    def dtype(self):
        """Return Numpy dtype of the object."""

        def array_dtype(x, template):
            if hasattr(template, "keys") and len(list(template.keys())) > 1:
                peek = x
                if isinstance(x, IterData):
                    peek = next(iter(x))
                return np.dtype(
                    [
                        (col, array_dtype(val, template[col]))
                        for col, val in zip(template.keys(), peek)
                    ]
                )
            else:
                return np.array(x).dtype

        return array_dtype(next(iter(self)), self.template)

    def iterdata(self):
        """Included for code symmetry with Types"""
        return iter(self)

    def __iter__(self):
        data = iter(self.stream)

        for f in self.ifilter:
            data = filter(f, data)
        for m in self.imap:
            data = map(m, data)
        for s in self.islice:
            data = itertools.islice(data, s.start, s.stop, s.step)

        return data

    def __copy__(self):
        """Return a lightweight copy of the object."""
        return IterData(
            self.stream,
            copy.copy(self.template),
            self.ifilter[:],
            self.imap[:],
            self.islice[:],
            self.level,
        )

    def __repr__(self):
        return "<IterData to stream %r>" % self.stream

    def __getitem__(self, key):
        out = copy.copy(self)

        # return a child, and adjust the data so that only the corresponding
        # column is returned
        if isinstance(key, str):
            try:
                col = list(self.template._all_keys()).index(key)
            except ValueError:
                raise KeyError(key)
            out.level += 1
            out.template = out.template[key]
            out.imap.append(deep_map(operator.itemgetter(col), out.level))

        # return a new sequence with the selected children
        elif isinstance(key, list):
            cols = [list(self.template.keys()).index(k) for k in key]
            out.template._visible_keys = key
            out.imap.append(
                deep_map(lambda row: tuple(row[i] for i in cols), out.level + 1)
            )

        # slice the data
        elif isinstance(key, (int, slice)):
            if isinstance(key, int):
                out.islice.append(slice(key, key + 1))
            else:
                out.islice.append(key)

        # filter the data; like slicing it, we can use ``itertools.ifilter``
        # only if the selection is applied to the outermost sequence, otherwise
        # we need to do a deep map
        elif isinstance(key, ConstraintExpression):
            f, m = build_filter(key, self.template)
            out.ifilter.append(f)
            out.imap.append(m)

        else:
            raise KeyError(key)

        return out

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            right = other.template.id
        else:
            right = encode(other)
        return ConstraintExpression("%s=%s" % (self.template.id, right))

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            right = other.template.id
        else:
            right = encode(other)
        return ConstraintExpression("%s!=%s" % (self.template.id, right))

    def __ge__(self, other):
        if isinstance(other, self.__class__):
            right = other.template.id
        else:
            right = encode(other)
        return ConstraintExpression("%s>=%s" % (self.template.id, right))

    def __le__(self, other):
        if isinstance(other, self.__class__):
            right = other.template.id
        else:
            right = encode(other)
        return ConstraintExpression("%s<=%s" % (self.template.id, right))

    def __gt__(self, other):
        if isinstance(other, self.__class__):
            right = other.template.id
        else:
            right = encode(other)
        return ConstraintExpression("%s>%s" % (self.template.id, right))

    def __lt__(self, other):
        if isinstance(other, self.__class__):
            right = other.template.id
        else:
            right = encode(other)
        return ConstraintExpression("%s<%s" % (self.template.id, right))


def fix_nested(template):
    """Apply ``IterData`` to nested sequences on iteration."""

    def func(row):
        return tuple(
            IterData(col, child) if isinstance(child, SequenceType) else col
            for col, child in zip(row, template.children())
        )

    return func


def deep_map(function, level):
    """Map a function inside a nested list, returning the modified data."""

    def out(row, level=level):
        if level == 1:
            return function(row)
        else:
            return [out(value, level - 1) for value in row]

    return out


def build_filter(expression, template):
    """Return a filter function based on a comparison expression."""
    id1, op, id2 = re.split("(<=|>=|!=|=~|>|<|=)", str(expression), 1)

    # calculate the column index were filtering and how deep it is
    try:
        id1 = id1[len(template.id) + 1 :]
        target = template
        for level, token in enumerate(id1.split(".")):
            parent1 = target.id
            keys = list(target._all_keys())
            col = keys.index(token)
            target = target[token]
        a = operator.itemgetter(col)
    except Exception:
        raise ConstraintExpressionError(
            'Invalid constraint expression: "{expression}" '
            '("{id}" is not a valid variable)'.format(expression=expression, id=id1)
        )

    # if we're comparing two variables they must be on the same sequence, so
    # ``parent1`` must be equal to ``parent2``
    if id2.rsplit(".", 1)[0] == parent1:  # parent2 == parent1
        keys = list(template._all_keys())
        col = keys.index(id2.split(".")[-1])
        b = operator.itemgetter(col)
    else:
        try:
            value = ast.literal_eval(id2)

            def b(row):
                return value

        except Exception:
            raise ConstraintExpressionError(
                'Invalid constraint expression: "{expression}" '
                '("{id}" is not valid)'.format(expression=expression, id=id2)
            )

    op = {
        "<": operator.lt,
        ">": operator.gt,
        "!=": operator.ne,
        "=": operator.eq,
        ">=": operator.ge,
        "<=": operator.le,
        "=~": lambda a, b: re.match(b, a),
    }[op]

    # if the filtering is applied in the outermost sequence we can simply pass
    # a filter, and ignore the map
    if level == 0:

        def f(row):
            return op(a(row), b(row))

        def m(row):
            return row

    # if the filtering is applied to a nested sequence we actually need to map
    # the outer data so that the inner data is filtered
    else:
        f = bool

        def recurse(row, tokens, target):
            token = tokens.pop(0)

            # return the filtered inner data
            if not tokens:
                return [col for col in row if op(a(col), b(col))]

            # navigate inside the sequence
            col = list(target.keys()).index(token)
            target = target[col]

            # modify data in place; we need to convert tuple to list
            row = list(row)
            row[col] = recurse(row[col], tokens, target)
            return tuple(row)

        def m(row):
            tokens = id1.split(".")
            return recurse(row, tokens, template)

    return f, m



"""Test DAS parsing functions."""

import unittest

import numpy as np

from pydap.parsers.das import add_attributes, parse_das
from pydap.parsers.dds import dds_to_dataset
from pydap.tests.test_parsers_dds import DDS

DAS = """Attributes {
    structure {
        b {
            Int value 1;
            Float32 missing nan;
            String foo "one", "two";
        }
    }
    structure.i {
        Int answer 42;
    }
    meta {
        String debug 1;
    }
    SPEH {
        Int debug 1;
        Int TIME 0;
        Float32 COADSX 1e20;
        String COADSY "zero";
    }
    floats {
        Float64 a nan;
        Float64 b -inf;
        Float64 c inf;
        Float64 d 17;
    }
}"""

# It is important to add attributes that have the same
# name as the dimensions of SPEH. This is an edge
# case that can break the das parser.


class TestParseDAS(unittest.TestCase):
    """Test DAS parser."""

    def setUp(self):
        """Load a dataset and apply DAS to it."""
        self.dataset = dds_to_dataset(DDS)
        attributes = parse_das(DAS)
        add_attributes(self.dataset, attributes)

    def test_basic(self):
        """Test a basic attribute."""
        self.assertEqual(self.dataset.structure.b.value, 1)

    def test_nan(self):
        """Test NaN."""
        self.assertTrue(np.isnan(self.dataset.structure.b.missing))

    def test_multiple_values(self):
        """Test attributes with multiple values."""
        self.assertEqual(self.dataset.structure.b.foo, ["one", "two"])

    def test_dot_attribute(self):
        """Test dotted attributes."""
        self.assertEqual(self.dataset.structure.i.answer, 42)

    def test_meta_attributes(self):
        """Test attributes not associated with any variables."""
        self.assertEqual(self.dataset.meta, {"debug": "1"})

    def test_SPEH_attributes(self):
        """Test attributes not associated with any variables."""
        self.assertEqual(self.dataset.SPEH.debug, 1)
        self.assertEqual(self.dataset.SPEH.attributes["TIME"], 0)
        self.assertEqual(self.dataset.SPEH.attributes["COADSX"], 1e20)
        self.assertEqual(self.dataset.SPEH.attributes["COADSY"], "zero")

    def test_float_attributes(self):
        """Test various values of float attributes."""
        self.assertTrue(np.isnan(self.dataset.floats["a"]))
        self.assertEqual(self.dataset.floats["b"], float("-inf"))
        self.assertEqual(self.dataset.floats["c"], float("inf"))
        self.assertEqual(self.dataset.floats["d"], 17.0)



"""Datasets for testing.

This module defines a few datasets for testing, covering the whole DAP data
model.

"""

import os
from collections import OrderedDict

import numpy as np

from pydap.client import open_file
from pydap.handlers.lib import IterData
from pydap.model import BaseType, DatasetType, GridType

# Note that DAP2 does not support signed bytes (signed 8bits integers).

# A very simple sequence: flat and with no strings. This sequence can be mapped
# directly to a Numpy structured array, and can be easily encoded and decoded
# in the DAP spec.
VerySimpleSequence = DatasetType("VerySimpleSequence")
VerySimpleSequence.createSequence("/sequence")
VerySimpleSequence.createVariable("/sequence.byte")
VerySimpleSequence.createVariable("/sequence.int")
VerySimpleSequence.createVariable("/sequence.float")
VerySimpleSequence["/sequence"].data = np.array(
    [
        (0, 1, 10.0),
        (1, 2, 20.0),
        (2, 3, 30.0),
        (3, 4, 40.0),
        (4, 5, 50.0),
        (5, 6, 60.0),
        (6, 7, 70.0),
        (7, 8, 80.0),
    ],
    dtype=[("byte", np.ubyte), ("int", "i4"), ("float", "f4")],
)


# A nested sequence.
NestedSequence = DatasetType("NestedSequence")
NestedSequence.createSequence("/location")
NestedSequence.createVariable("/location.lat")
NestedSequence.createVariable("/location.lon")
NestedSequence.createVariable("/location.elev")
NestedSequence.createSequence("/location.time_series")
NestedSequence.createVariable("/location.time_series.time")
NestedSequence.createVariable("/location.time_series.slp")
NestedSequence.createVariable("/location.time_series.wind")
NestedSequence["/location"].data = IterData(
    [
        (1, 1, 1, [(10, 11, 12), (21, 22, 23)]),
        (2, 4, 4, [(15, 16, 17)]),
        (3, 6, 9, []),
        (4, 8, 16, [(31, 32, 33), (41, 42, 43), (51, 52, 53), (61, 62, 63)]),
    ],
    NestedSequence.location,
)


# A simple array with bytes, strings and shorts. These types require special
# encoding for the DODS response.
SimpleArray = DatasetType("SimpleArray")
SimpleArray.createVariable("byte", data=np.arange(5, dtype=np.ubyte))
SimpleArray.createVariable("string", data=np.array(["one", "two"]))
SimpleArray.createVariable("short", data=np.array(1, dtype="h"))


DODS = os.path.join(os.path.dirname(__file__), "data/rainfall_time_malaysia.cdp.dods")
DAS = os.path.join(os.path.dirname(__file__), "data/rainfall_time_malaysia.cdp.das")
dapper = open_file(DODS, DAS)


# dataset from http://test.opendap.org:8080/dods/dts/D1.asc
D1 = DatasetType("EOSDB.DBO", type="Drifters")
D1.createSequence("/Drifters")
D1.createVariable("/Drifters.instrument_id")
D1.createVariable("/Drifters.location")
D1.createVariable("/Drifters.latitude")
D1.createVariable("/Drifters.longitude")
D1.Drifters.data = np.array(
    np.rec.fromrecords(
        list(
            zip(
                [
                    "This is a data test string (pass {0}).".format(1 + i * 2)
                    for i in range(5)
                ],
                [
                    "This is a data test string (pass {0}).".format(i * 2)
                    for i in range(5)
                ],
                [1000.0, 999.95, 999.80, 999.55, 999.20],
                [999.95, 999.55, 998.75, 997.55, 995.95],
            )
        ),
        names=list(D1.Drifters.keys()),
    )
)


# testing structures
SimpleStructure = DatasetType("SimpleStructure")
SimpleStructure.createStructure(
    name="/types",
    key="value",
    nested=OrderedDict(
        [
            ("string", "bar"),
            ("list", [42, 43]),
            ("array", np.array(1)),
            ("float", 1000.0),
        ]
    ),
)
SimpleStructure.createVariable(name="/types.b", data=np.array(-10, np.byte))
SimpleStructure.createVariable(name="/types.ub", data=np.array(10, np.ubyte))
SimpleStructure.createVariable(name="/types.i32", data=np.array(-10, np.int32))
SimpleStructure.createVariable(name="/types.ui32", data=np.array(10, np.uint32))
SimpleStructure.createVariable(name="/types.i16", data=np.array(-10, np.int16))
SimpleStructure.createVariable(name="/types.ui16", data=np.array(10, np.uint16))
SimpleStructure.createVariable(name="/types.f32", data=np.array(100.0, np.float32))
SimpleStructure.createVariable(name="/types.f64", data=np.array(1000.0, np.float64))
SimpleStructure.createVariable(
    name="/types.s", data=np.array("This is a data test string (pass 0).")
)
SimpleStructure.createVariable(name="/types.u", data=np.array("http://www.dods.org"))
SimpleStructure.createVariable(name="/types.U", data=np.array("test unicode", str))


# test grid
rain = DatasetType("test")
rain["rain"] = GridType("rain")
rain["rain"]["rain"] = BaseType(
    "rain", np.arange(6).reshape(2, 3), dimensions=("y", "x")
)
rain["rain"]["x"] = BaseType("x", np.arange(3), units="degrees_east")
rain["rain"]["y"] = BaseType("y", np.arange(2), units="degrees_north")


# test for ``bounds`` function
bounds = DatasetType("test")
bounds.createSequence("/sequence")
bounds.createVariable("/sequence.lon", axis="X")
bounds.createVariable("/sequence.lat", axis="Y")
bounds.createVariable("/sequence.depth", axis="Z")
bounds.createVariable("/sequence.time", axis="T", units="days since 1970-01-01")
bounds["sequence"]["measurement"] = BaseType("measurement")
bounds.sequence.data = np.array(
    np.rec.fromrecords(
        [
            (100, -10, 0, -1, 42),
            (200, 10, 500, 1, 43),
        ],
        names=list(bounds.sequence.keys()),
    )
)


# test for density
ctd = DatasetType("ctd")
ctd.createSequence("cast")
ctd.createVariable("/cast.temperature")
ctd.createVariable("/cast.salinity")
ctd.createVariable("/cast.pressure")

ctd.cast.data = np.array(
    np.rec.fromrecords(
        [
            (21, 35, 0),
            (15, 35, 100),
        ],
        names=list(ctd.cast.keys()),
    )
)

# a simple sequence, simulating a CTD profile
SimpleSequence = DatasetType(
    "SimpleSequence", description="A simple sequence for testing.", nested={"value": 42}
)
SimpleSequence.createSequence("/cast")
SimpleSequence.createVariable("/cast.id")
SimpleSequence.createVariable("/cast.lon", axis="X")
SimpleSequence.createVariable("/cast.lat", axis="Y")
SimpleSequence.createVariable("/cast.depth", axis="Z")
SimpleSequence.createVariable("/cast.time", axis="T", units="days since 1970-01-01")
SimpleSequence.createVariable("/cast.temperature")
SimpleSequence.createVariable("/cast.salinity")
SimpleSequence.createVariable("/cast.pressure")

SimpleSequence["cast"].data = np.array(
    np.rec.fromrecords(
        [
            ("1", 100, -10, 0, -1, 21, 35, 0),
            ("2", 200, 10, 500, 1, 15, 35, 100),
        ],
        names=list(SimpleSequence["cast"].keys()),
    )
)
# SimpleSequence["cast"].data = IterData([
#     ("1", 100, -10,   0, -1, 21, 35,   0),
#     ("2", 200,  10, 500,  1, 15, 35, 100),
# ], SimpleSequence.cast)

# a simple grid
SimpleGrid = DatasetType("SimpleGrid", description="A simple grid for testing.")
SimpleGrid["SimpleGrid"] = GridType("SimpleGrid")
SimpleGrid["SimpleGrid"]["SimpleGrid"] = BaseType(
    "SimpleGrid", np.arange(6).reshape(2, 3), dimensions=("y", "x")
)
SimpleGrid["x"] = SimpleGrid["SimpleGrid"]["x"] = BaseType(
    "x", np.arange(3), axis="X", units="degrees_east"
)
SimpleGrid["y"] = SimpleGrid["SimpleGrid"]["y"] = BaseType(
    "y", np.arange(2), axis="Y", units="degrees_north"
)

SimpleGroup = DatasetType(
    "example dataset",
    Description="A simple group for testing.",
    dimensions={"time": 1, "nv": 2},
)
SimpleGroup.createGroup(
    "SimpleGroup",
    dimensions={"Y": 4, "X": 4},
    Description="Test group with numerical data",
)
SimpleGroup.createVariable(
    name="/SimpleGroup/Temperature",
    data=np.arange(10, 26, 1, dtype="f4").reshape(1, 4, 4),
    units="degrees_celsius",
    dimensions=("/time", "/SimpleGroup/Y", "/SimpleGroup/X"),
    _FillValue=np.inf,
)
SimpleGroup.createVariable(
    name="/SimpleGroup/Salinity",
    data=30 * np.ones(16, dtype="f4").reshape(1, 4, 4),
    units="psu",
    dimensions=("/time", "/SimpleGroup/Y", "/SimpleGroup/X"),
    _FillValue=np.nan,
)
SimpleGroup.createVariable(
    name="/SimpleGroup/Y", data=np.arange(4, dtype="i2"), dimensions=("/SimpleGroup/Y",)
)
SimpleGroup.createVariable(
    name="/SimpleGroup/X", data=np.arange(4, dtype="i2"), dimensions=("/SimpleGroup/X",)
)
SimpleGroup.createVariable(
    name="/time",
    data=np.array(0.5, dtype="f4"),
    dimensions=("/time",),
    attributes={
        "standard_name": "time",
        "bounds": "time_bnds",
    },
)
SimpleGroup.createVariable(
    name="/time_bnds", data=np.arange(2, dtype="f4"), dimensions=("/time", "/nv")
)


# a faulty grid
FaultyGrid = DatasetType("FaultyGrid", description="A faulty grid for testing.")
FaultyGrid["FaultyGrid"] = GridType("FaultyGrid")
FaultyGrid["FaultyGrid"]["FaultyGrid"] = BaseType(
    "FaultyGrid", np.arange(6).reshape(2, 3), dimensions=("y", "x")
)
FaultyGrid["x"] = FaultyGrid["FaultyGrid"]["x"] = BaseType(
    "x", np.arange(3), axis="X", code=1
)
FaultyGrid["y"] = FaultyGrid["FaultyGrid"]["y"] = BaseType(
    "y", np.arange(2), axis="Y", code=np.int32([])
)



import os

import pytest
import requests

import pydap.net
from pydap.cas import urs
from pydap.client import open_url

url = (
    "https://goldsmr3.gesdisc.eosdis.nasa.gov/opendap/"
    "MERRA_MONTHLY/MAIMCPASM.5.2.0/1979/"
    "MERRA100.prod.assim.instM_3d_asm_Cp.197901.hdf"
)
test_url = url + ".dods?SLP[0:1:0][0:1:10][0:1:10]"
test_url_2 = url + ".dods?PS[0:1:0][0:1:10][0:1:10]"


@pytest.mark.auth
@pytest.mark.prod_url
@pytest.mark.skipif(
    not (os.environ.get("USERNAME_URS") and os.environ.get("PASSWORD_URS")),
    reason=("Without auth credentials, " "this test cannot work"),
)
def test_basic_urs_auth():
    """
    Set up PyDAP to use the URS request() function.

    The intent here is to ensure that pydap.net is able to
    open and url if and only if requests is able to
    open the same url.
    """
    session = urs.setup_session(
        os.environ.get("USERNAME_URS"), os.environ.get("PASSWORD_URS"), check_url=url
    )

    # Check that the requests library can access the link:
    res = requests.get(test_url, cookies=session.cookies)
    assert res.status_code == 200
    res.close()

    # Check that the pydap library can access the link:
    res = pydap.net.follow_redirect(test_url, session=session)
    assert res.status_code == 200

    # Check that the pydap library can access another link:
    res = pydap.net.follow_redirect(test_url_2, session=session)
    assert res.status_code == 200
    session.close()


@pytest.mark.auth
@pytest.mark.prod_url
@pytest.mark.skipif(
    not (os.environ.get("USERNAME_URS") and os.environ.get("PASSWORD_URS")),
    reason=("Without auth credentials, " "this test cannot work"),
)
def test_basic_urs_query():
    session = urs.setup_session(
        os.environ.get("USERNAME_URS"), os.environ.get("PASSWORD_URS"), check_url=url
    )
    # Ensure authentication:
    res = pydap.net.follow_redirect(test_url, session=session)
    assert res.status_code == 200
    dataset = open_url(url, session=session)
    expected_data = [
        [
            [99066.15625, 99066.15625, 99066.15625, 99066.15625, 99066.15625],
            [98868.15625, 98870.15625, 98872.15625, 98874.15625, 98874.15625],
            [98798.15625, 98810.15625, 98820.15625, 98832.15625, 98844.15625],
            [98856.15625, 98828.15625, 98756.15625, 98710.15625, 98776.15625],
            [99070.15625, 99098.15625, 99048.15625, 98984.15625, 99032.15625],
        ]
    ]
    assert (dataset["SLP"][0, :5, :5] == expected_data).all()
    session.close()



"""Test the base functions for handlers."""

import unittest

from webob import Request

import pydap.responses
from pydap.model import DatasetType
from pydap.responses.lib import BaseResponse, load_responses
from pydap.tests.datasets import VerySimpleSequence


class TestLoadResponse(unittest.TestCase):
    """Test that responses are loaded correctly."""

    def setUp(self):
        """Load all available responses.

        At least the DDS, DAS, DODS, DMR, HTML, ASC and version responses should be
        loaded, since they are included in pydap. Other third-party responses
        may also be present.

        """
        self.responses = load_responses()

    def test_responses_das(self):
        """Test that the DAS response is loaded."""
        self.assertIs(self.responses["das"], pydap.responses.das.DASResponse)

    def test_responses_dds(self):
        """Test that the DDS response is loaded."""
        self.assertIs(self.responses["dds"], pydap.responses.dds.DDSResponse)

    def test_responses_dods(self):
        """Test that the DODS response is loaded."""
        self.assertIs(self.responses["dods"], pydap.responses.dods.DODSResponse)

    def test_responses_html(self):
        """Test that the HTML response is loaded."""
        self.assertIs(self.responses["html"], pydap.responses.html.HTMLResponse)

    def test_responses_ascii(self):
        """Test that the ASCII response is loaded."""
        self.assertIs(self.responses["ascii"], pydap.responses.ascii.ASCIIResponse)
        self.assertIs(self.responses["asc"], pydap.responses.ascii.ASCIIResponse)

    def test_responses_dmr(self):
        """Test that the dmr response is loaded."""
        self.assertIs(self.responses["dmr"], pydap.responses.dmr.DMRResponse)

    def test_responses_ver(self):
        """Test that the version response is loaded."""
        self.assertIs(self.responses["ver"], pydap.responses.version.VersionResponse)


class TestBaseResponse(unittest.TestCase):
    """Test the response super class."""

    def setUp(self):
        """Instantiate a response."""
        self.response = BaseResponse(VerySimpleSequence)

    def test_call(self):
        """Test calling the WSGI app."""
        req = Request.blank("/")
        res = req.get_response(self.response)
        self.assertEqual(res.status, "200 OK")

        # note that the iterable returned by our WSGI app is the object itself:
        self.assertIs(res.app_iter, self.response)

    def test_serialization(self):
        """Test the serialization code."""
        req = Request.blank("/")
        res = req.get_response(self.response)

        # get a dataset when we pass ``DatsetType``
        self.assertIs(
            res.app_iter.x_wsgiorg_parsed_response(DatasetType), VerySimpleSequence
        )

        # when we pass something the method should return None
        self.assertIsNone(res.app_iter.x_wsgiorg_parsed_response("dataset"))

    def test_iter(self):
        """Test that calling the base class directly raises an exception."""
        with self.assertRaises(NotImplementedError):
            iter(self.response)



"""pydap error response.

pydap will capture exceptions returned by the system and return them properly
formated as a DAP error response.

"""

from io import StringIO
from traceback import print_exception

from webob import Response

from ..lib import __version__, encode


class ErrorResponse(object):
    """A specialized response for errors.

    This is a special response used when an exception is captured and passed to
    the user as an Opendap error message:

    """

    def __init__(self, info):
        # get exception message
        buf = StringIO()
        print_exception(*info, file=buf)
        message = encode(buf.getvalue())

        # build error message
        code = getattr(info[0], "code", -1)
        self.body = str(
            """Error {{
    code = {0};
    message = {1};
}}"""
        ).format(code, message)

    def __call__(self, environ, start_response):
        res = Response()
        res.text = self.body
        res.status = "500 Internal Error"
        res.content_type = "text/plain"
        res.charset = "utf-8"
        res.headers.add("Content-description", "OPeNDAP_error")
        res.headers.add("OPeNDAP-Server", "pydap/%s" % __version__)

        return res(environ, start_response)



"""Test the basic DAP functions."""

import unittest
from sys import maxsize as MAXSIZE

import numpy as np

from pydap.exceptions import ConstraintExpressionError
from pydap.lib import (
    _quote,
    combine_slices,
    encode,
    fix_shorthand,
    fix_slice,
    get_var,
    hyperslab,
    tree,
    walk,
)
from pydap.model import BaseType, DatasetType, SequenceType, StructureType


class TestQuote(unittest.TestCase):
    """Test quoting.

    According to the DAP 2 specification a variable name MUST contain only
    upper or lower case letters, numbers, or characters from the set

        _ ! ~ * ' - "

    All other characters must be escaped. This includes the period, which is
    normally not quoted by ``urllib.quote``.

    """

    def test_quoting(self):
        """Test a simple quoting."""
        self.assertEqual(_quote("White space"), "White%20space")

    def test_quoting_period(self):
        """Test if period is also quoted."""
        self.assertEqual(_quote("Period."), "Period%2E")


class TestEncode(unittest.TestCase):
    """Test encoding.

    According to the DAP 2 specification, numbers must be encoded using the C
    notation "%.6g". Other objects are encoded as escaped strings.

    """

    def test_integer(self):
        """Test integer encoding."""
        self.assertEqual(encode(1), "1")

    def test_float(self):
        """Test floating encoding."""
        self.assertEqual(encode(np.pi), "3.14159")

    def test_string(self):
        """Test string encoding."""
        self.assertEqual(encode("test"), '"test"')

    def test_string_with_quotation(self):
        """Test encoding a string with a quotation mark."""
        self.assertEqual(encode('this is a "test"'), '"this is a "test""')

    def test_unicode(self):
        """Unicode objects are encoded just like strings."""
        self.assertEqual(encode("test"), '"test"')

    def test_obj(self):
        """Other objects are encoded according to their ``repr``."""
        self.assertEqual(encode({}), '"{}"')

    def test_numpy_string(self):
        self.assertEqual(encode(np.array("1", dtype="<U1")), '"1"')

    def test_numpy_ndim_gt_0(self):
        # test a numpy array with ndim > 0
        # associated with Deprecation warning numpy > 1.25
        # see GH issue https://github.com/pydap/pydap/issues/319
        # also PR https://github.com/pydap/pydap/pull/343
        array = np.array([(2.300110099991, 4.0)])
        self.assertEqual(encode(array), '"[[2.300110 4.000000]]"')


class TestFixSlice(unittest.TestCase):
    """Test the ``fix_slice`` function."""

    def test_not_tuple(self):
        """Non tuples should be converted and handled correctly."""
        x = np.arange(10)

        slice1 = 0
        slice2 = fix_slice(slice1, x.shape)

        # ``fix_slice`` will convert to a tuple
        self.assertEqual(slice2, (0,))

        # assert that the slice is equivalent to the original
        np.testing.assert_array_equal(x[slice1], x[slice2])

    def test_ellipsis(self):
        """Expand Ellipsis to occupy the missing dimensions."""
        x = np.arange(6).reshape(2, 3, 1)

        slice1 = Ellipsis, 0
        slice2 = fix_slice(slice1, x.shape)

        # an Ellipsis is expanded to slice(None)
        self.assertEqual(slice2, ((slice(0, 2, 1), slice(0, 3, 1), 0)))
        np.testing.assert_array_equal(x[slice1], x[slice2])

    def test_negative_int(self):
        """Negative values are converted to positive."""
        x = np.arange(10)

        slice1 = -5
        slice2 = fix_slice(slice1, x.shape)

        self.assertEqual(slice2, (5,))
        np.testing.assert_array_equal(x[slice1], x[slice2])

    def test_negative_start(self):
        """Test for slices with a negative start."""
        x = np.arange(10)

        slice1 = slice(-8, 8)
        slice2 = fix_slice(slice1, x.shape)

        self.assertEqual(slice2, (slice(2, 8, 1),))
        np.testing.assert_array_equal(x[slice1], x[slice2])

    def test_negative_stop(self):
        """Test for slices with a negative stop."""
        x = np.arange(10)

        slice1 = slice(2, -2)
        slice2 = fix_slice(slice1, x.shape)

        self.assertEqual(slice2, (slice(2, 8, 1),))
        np.testing.assert_array_equal(x[slice1], x[slice2])

    def test_non_zero_slicestart(self):
        x = np.arange(0, 100)
        slice1 = slice(100, 200)
        slice2 = fix_slice(slice1, x.shape)
        self.assertEqual(slice2, (slice(100, 200, 1),))


class TestCombineSlices(unittest.TestCase):
    """Test the ``combine_slices`` function."""

    def test_not_tuple(self):
        """The function fails when one of the slices is not a tuple."""
        slice1 = 0
        slice2 = (0,)
        with self.assertRaises(TypeError):
            combine_slices(slice1, slice2)
        with self.assertRaises(TypeError):
            combine_slices(slice2, slice1)

    def test_integer(self):
        """Test slices that are just integers."""
        slice1 = (0,)
        slice2 = (1,)
        combined = combine_slices(slice1, slice2)
        self.assertEqual(combined, (slice(1, 1, 1),))

    def test_stops_none(self):
        """Test when both of the slices have ``None`` for stop."""
        x = np.arange(10)
        slice1 = (slice(0, None),)
        slice2 = (slice(5, None),)
        combined = combine_slices(slice1, slice2)
        self.assertEqual(combined, (slice(5, None, 1),))
        np.testing.assert_array_equal(x[combined], x[slice1][slice2])

    def test_first_stop_none(self):
        """Test when the first slice has ``None`` for stop."""
        x = np.arange(10)
        slice1 = (slice(5, None),)
        slice2 = (slice(0, 8),)
        combined = combine_slices(slice1, slice2)
        self.assertEqual(combined, (slice(5, 13, 1),))
        np.testing.assert_array_equal(x[combined], x[slice1][slice2])

    def test_second_stop_none(self):
        """Test when the second slice has ``None`` for stop."""
        x = np.arange(10)
        slice1 = (slice(0, 8),)
        slice2 = (slice(5, None),)
        combined = combine_slices(slice1, slice2)
        self.assertEqual(combined, (slice(5, 8, 1),))
        np.testing.assert_array_equal(x[combined], x[slice1][slice2])

    def test_all_values(self):
        """Test when start and stop are all integers."""
        x = np.arange(20)
        slice1 = (slice(0, 8),)
        slice2 = (slice(5, 6),)
        combined = combine_slices(slice1, slice2)
        self.assertEqual(combined, (slice(5, 6, 1),))
        np.testing.assert_array_equal(x[combined], x[slice1][slice2])


class TestHyperslab(unittest.TestCase):
    """Test hyperslab generation from Python slices."""

    def test_no_tuple(self):
        """Test that slices that are not tuples work."""
        slice_ = slice(0)
        self.assertEqual(hyperslab(slice_), "[0:1:%d]" % (MAXSIZE - 1))

    def test_remove(self):
        """Test that excess slices are removed."""
        slice_ = (slice(0), slice(None))
        self.assertEqual(hyperslab(slice_), "[0:1:%d]" % (MAXSIZE - 1))

    def test_ndimensional(self):
        """Test n-dimensions slices."""
        slice_ = (slice(1, 10, 1), slice(2, 10, 2))
        self.assertEqual(hyperslab(slice_), "[1:1:9][2:2:9]")


class TestWalk(unittest.TestCase):
    """Test the ``walk`` function to iterate over a dataset."""

    def setUp(self):
        """Create a basic dataset."""
        self.dataset = DatasetType("a")
        self.dataset["b"] = BaseType("b")
        self.dataset["c"] = StructureType("c")
        self.dataset["d"] = SequenceType("d")

    def test_walk(self):
        """Test that all variables are yielded."""
        self.assertEqual(
            list(walk(self.dataset)),
            [self.dataset, self.dataset.b, self.dataset.c, self.dataset.d],
        )

    def test_walk_type(self):
        """Test the filtering of variables yielded."""
        self.assertEqual(list(walk(self.dataset, BaseType)), [self.dataset.b])
        self.assertEqual(list(walk(self.dataset, SequenceType)), [self.dataset.d])


class TestTree(unittest.TestCase):
    """Test the ``tree`` func"""

    def setUp(self):
        self.dataset = DatasetType("name")
        self.dataset["a"] = BaseType("a")
        self.dataset.createGroup("Group1")
        self.dataset.createSequence("/Group1/Seq1")
        self.dataset.createVariable("/Group1/Seq1/b")
        self.dataset.createVariable("/Group1/Seq1/c")
        self.dataset["d"] = StructureType("d")
        self.dataset["e"] = SequenceType("e")

    def test_tree_repr(self):
        self.assertEqual(self.dataset.tree(), tree(self.dataset))
        self.assertEqual(self.dataset["Group1"].tree(), tree(self.dataset["Group1"]))


class TestFixShorthand(unittest.TestCase):
    """Test the ``fix_shorthand`` function."""

    def test_fix_projection(self):
        """Test a dataset that can use the shorthand notation."""
        dataset = DatasetType("a")
        dataset["b"] = StructureType("b")
        dataset["b"]["c"] = BaseType("c")

        projection = [[("c", ())]]
        self.assertEqual(fix_shorthand(projection, dataset), [[("b", ()), ("c", ())]])

    def test_conflict(self):
        """Test a dataset with conflicting short names."""
        dataset = DatasetType("a")
        dataset["b"] = StructureType("b")
        dataset["b"]["c"] = BaseType("c")
        dataset["d"] = StructureType("d")
        dataset["d"]["c"] = BaseType("c")

        projection = [[("c", ())]]
        with self.assertRaises(ConstraintExpressionError):
            fix_shorthand(projection, dataset)


class TestGetVar(unittest.TestCase):
    """Test the ``get_var`` function."""

    def test_get_var(self):
        """Test that the id is returned properly."""
        dataset = DatasetType("a")
        dataset["b"] = StructureType("b")
        dataset["b"]["c"] = BaseType("c")

        self.assertEqual(get_var(dataset, "b.c"), dataset["b"]["c"])



"""Test DDS parsing functions."""

import os
import unittest

import numpy as np

from pydap.model import BaseType, StructureType
from pydap.parsers.dds import dds_to_dataset


def load_dds_file(file_path):
    abs_path = os.path.join(os.path.dirname(__file__), file_path)
    with open(abs_path, "r") as dds_file:
        dds = dds_file.read()
    dataset = dds_to_dataset(dds)
    return dataset


# DDS_FLAT = os.path.join(os.path.dirname(__file__), "data/flatgroup.dds")


DDS = """Dataset {
    Structure {
        Byte b;
        Int i;
        Uint ui;
        Int16 i16;
        Uint16 ui16;
        Int32 i32;
        Uint32 ui32;
        Float32 f32;
        Float64 f64;
        String s;
        Url u;
    } structure;
    Sequence {
        Int32 a;
    } sequence;
    Int32 b[10];
    Int32 c[c = 10];
    Grid {
        Array:
            Float32 SPEH[TIME = 12][COADSY = 90][COADSX = 180];
        Maps:
            Float64 TIME[TIME = 12];
            Float64 COADSY[COADSY = 90];
            Float64 COADSX[COADSX = 180];
    } SPEH;
} d;"""


class TestBuildDataset(unittest.TestCase):
    """Test DDS parser."""

    def setUp(self):
        """Parse the whole dataset."""
        self.dataset = dds_to_dataset(DDS)
        self.dataset2 = load_dds_file("data/flatgroup.dds")

    def test_structure(self):
        """Test the structure."""
        self.assertIsInstance(self.dataset.structure, StructureType)

    def test_byte(self):
        """Test byte parsing."""
        self.assertIsInstance(self.dataset.structure.b, BaseType)
        self.assertEqual(self.dataset.structure.b.dtype, np.dtype("B"))
        self.assertEqual(self.dataset.structure.b.shape, ())

    def test_integer(self):
        """Test integer parsing."""
        self.assertIsInstance(self.dataset.structure.i, BaseType)
        self.assertEqual(self.dataset.structure.i.dtype, np.dtype(">i"))
        self.assertEqual(self.dataset.structure.i.shape, ())

    def test_unsigned_integer(self):
        """Test unsigned integer parsing."""
        self.assertIsInstance(self.dataset.structure.ui, BaseType)
        self.assertEqual(self.dataset.structure.ui.dtype, np.dtype(">I"))
        self.assertEqual(self.dataset.structure.ui.shape, ())

    def test_integer_16(self):
        """Test integer 16 parsing."""
        self.assertIsInstance(self.dataset.structure.i16, BaseType)
        self.assertEqual(self.dataset.structure.i16.dtype, np.dtype(">h"))
        self.assertEqual(self.dataset.structure.i16.shape, ())

    def test_unsigned_integer_16(self):
        """Test unsigned integer 16 parsing."""
        self.assertIsInstance(self.dataset.structure.ui16, BaseType)
        self.assertEqual(self.dataset.structure.ui16.dtype, np.dtype(">H"))
        self.assertEqual(self.dataset.structure.ui16.shape, ())

    def test_integer_32(self):
        """Test integer 32 parsing."""
        self.assertIsInstance(self.dataset.structure.i32, BaseType)
        self.assertEqual(self.dataset.structure.i32.dtype, np.dtype(">i"))
        self.assertEqual(self.dataset.structure.i32.shape, ())

    def test_unsigned_integer_32(self):
        """Test unsigned integer 32 parsing."""
        self.assertIsInstance(self.dataset.structure.ui32, BaseType)
        self.assertEqual(self.dataset.structure.ui32.dtype, np.dtype(">I"))
        self.assertEqual(self.dataset.structure.ui32.shape, ())

    def test_float_32(self):
        """Test float 32 parsing."""
        self.assertIsInstance(self.dataset.structure.f32, BaseType)
        self.assertEqual(self.dataset.structure.f32.dtype, np.dtype(">f"))
        self.assertEqual(self.dataset.structure.f32.shape, ())

    def test_float_64(self):
        """Test float 64 parsing."""
        self.assertIsInstance(self.dataset.structure.f64, BaseType)
        self.assertEqual(self.dataset.structure.f64.dtype, np.dtype(">d"))
        self.assertEqual(self.dataset.structure.f64.shape, ())

    def test_string(self):
        """Test string parsing."""
        self.assertIsInstance(self.dataset.structure.s, BaseType)
        self.assertEqual(self.dataset.structure.s.dtype, np.dtype("|S128"))
        self.assertEqual(self.dataset.structure.s.shape, ())

    def test_flatgroup(self):
        self.assertEqual(self.dataset2.groups(), {})
        self.assertEqual(
            list(self.dataset2.keys()),
            [
                "/A/B/sst",
                "/A/B/lat",
                "/A/B/lon",
                "/A/B/time",
                "/A/B/time_bnds",
                "/A/B/Vertical_binsize",
            ],
        )

    def test_url(self):
        """Test url parsing."""
        self.assertIsInstance(self.dataset.structure.u, BaseType)
        self.assertEqual(self.dataset.structure.u.dtype, np.dtype("|S128"))
        self.assertEqual(self.dataset.structure.u.shape, ())



"""Test the DODS response."""

import copy
import unittest
from collections import OrderedDict

import numpy as np
from numpy.lib import Arrayterator
from webob.headers import ResponseHeaders
from webtest import TestApp as App

from pydap.handlers.lib import BaseHandler
from pydap.lib import END_OF_SEQUENCE, START_OF_SEQUENCE, __version__
from pydap.responses.dods import DODSResponse, dods
from pydap.tests.datasets import (
    NestedSequence,
    SimpleArray,
    SimpleGrid,
    SimpleSequence,
    SimpleStructure,
    VerySimpleSequence,
)


class TestDODSResponse(unittest.TestCase):
    """Test the DODS response."""

    def setUp(self):
        """Create a simple WSGI app."""
        app = App(BaseHandler(VerySimpleSequence))
        self.res = app.get("/.dods")

    def test_dispatcher(self):
        """Test that the dispatcher works on any object."""
        with self.assertRaises(StopIteration):
            dods(None)

    def test_status(self):
        """Test the status code."""
        self.assertEqual(self.res.status, "200 OK")

    def test_content_type(self):
        """Test the content type."""
        self.assertEqual(self.res.content_type, "application/octet-stream")

    def test_charset(self):
        """Test the charset."""
        self.assertEqual(self.res.charset, None)

    def test_headers(self):
        """Test the headers in the response."""
        self.assertEqual(
            self.res.headers,
            ResponseHeaders(
                [
                    ("OPeNDAP-Server", "pydap/" + __version__),
                    ("Content-description", "dods_data"),
                    ("Content-type", "application/octet-stream"),
                    ("Access-Control-Allow-Origin", "*"),
                    (
                        "Access-Control-Allow-Headers",
                        "Origin, X-Requested-With, Content-Type",
                    ),
                    ("Content-Length", "238"),
                ]
            ),
        )

    def test_body(self):
        """Test response body."""
        dds, xdrdata = self.res.body.split(b"\nData:\n", 1)
        dds = dds.decode("ascii")
        self.assertEqual(
            dds,
            """Dataset {
    Sequence {
        Byte byte;
        Int32 int;
        Float32 float;
    } sequence;
} VerySimpleSequence;""",
        )

        self.assertEqual(
            xdrdata,
            START_OF_SEQUENCE + b"\x00"
            b"\x00\x00\x00\x01"
            b"A \x00\x00" + START_OF_SEQUENCE + b"\x01"
            b"\x00\x00\x00\x02"
            b"A\xa0\x00\x00" + START_OF_SEQUENCE + b"\x02"
            b"\x00\x00\x00\x03"
            b"A\xf0\x00\x00" + START_OF_SEQUENCE + b"\x03"
            b"\x00\x00\x00\x04"
            b"B \x00\x00" + START_OF_SEQUENCE + b"\x04"
            b"\x00\x00\x00\x05"
            b"BH\x00\x00" + START_OF_SEQUENCE + b"\x05"
            b"\x00\x00\x00\x06"
            b"Bp\x00\x00" + START_OF_SEQUENCE + b"\x06"
            b"\x00\x00\x00\x07"
            b"B\x8c\x00\x00" + START_OF_SEQUENCE + b"\x07"
            b"\x00\x00\x00\x08"
            b"B\xa0\x00\x00" + END_OF_SEQUENCE,
        )


class TestDODSResponseGrid(unittest.TestCase):
    """Test the DODS response with a grid."""

    def setUp(self):
        """Create a simple WSGI app."""
        app = App(BaseHandler(SimpleGrid))
        self.res = app.get("/.dods")

    def test_body(self):
        """Test the response body."""
        dds, xdrdata = self.res.body.split(b"\nData:\n", 1)
        dds = dds.decode("ascii")
        self.assertEqual(
            dds,
            """Dataset {
    Grid {
        Array:
            Int32 SimpleGrid[y = 2][x = 3];
        Maps:
            Int32 x[x = 3];
            Int32 y[y = 2];
    } SimpleGrid;
    Int32 x[x = 3];
    Int32 y[y = 2];
} SimpleGrid;""",
        )

        self.assertEqual(
            xdrdata,
            b"\x00\x00\x00\x06"  # length
            b"\x00\x00\x00\x06"  # length, again
            b"\x00\x00\x00\x00"
            b"\x00\x00\x00\x01"
            b"\x00\x00\x00\x02"
            b"\x00\x00\x00\x03"
            b"\x00\x00\x00\x04"
            b"\x00\x00\x00\x05"
            b"\x00\x00\x00\x03"
            b"\x00\x00\x00\x03"
            b"\x00\x00\x00\x00"
            b"\x00\x00\x00\x01"
            b"\x00\x00\x00\x02"
            b"\x00\x00\x00\x02"
            b"\x00\x00\x00\x02"
            b"\x00\x00\x00\x00"
            b"\x00\x00\x00\x01"
            b"\x00\x00\x00\x03"
            b"\x00\x00\x00\x03"
            b"\x00\x00\x00\x00"
            b"\x00\x00\x00\x01"
            b"\x00\x00\x00\x02"
            b"\x00\x00\x00\x02"
            b"\x00\x00\x00\x02"
            b"\x00\x00\x00\x00"
            b"\x00\x00\x00\x01",
        )


class TestDODSResponseStructure(unittest.TestCase):
    """Test the DODS response with a sequence that has a string."""

    def setUp(self):
        """Create a simple WSGI app."""
        app = App(BaseHandler(SimpleStructure))
        self.res = app.get("/.dods")

    def test_body(self):
        """Test response body."""
        arr = np.array(-10, dtype=np.int16).astype(">i")
        arr_1d = np.array([arr])
        newbyteo = arr_1d.dtype.newbyteorder(">")
        newarr = arr_1d.astype(newbyteo)
        # View with new byte order
        expected_result = newarr.view(newbyteo)

        # begins here
        dds, xdrdata = self.res.body.split(b"\nData:\n", 1)
        dds = dds.decode("ascii")
        self.assertEqual(
            dds,
            """Dataset {
    Structure {
        Int16 b;
        Byte ub;
        Int32 i32;
        UInt32 ui32;
        Int16 i16;
        UInt16 ui16;
        Float32 f32;
        Float64 f64;
        String s;
        String u;
        String U;
    } types;
} SimpleStructure;""",
        )

        expected = OrderedDict(
            [
                # -10 signed byte upconv. to Int32 for transfer
                ("b", (np.array(-10, dtype=np.byte).astype(">i").tobytes())),
                # 10 usigned byte padded to multiple of 4 bytes for transfer
                ("ub", (np.array(10, dtype=np.ubyte).tobytes()) + b"\x00\x00\x00"),
                ("i32", (np.array(-10, dtype=np.int32).astype(">i").tobytes())),
                ("ui32", (np.array(10, dtype=np.uint32).astype(">I").tobytes())),
                # -10 int16 upconv. to int32 for transfer
                ("i16", expected_result.tobytes()),
                # 10 uint16 upconv. to uint32 for transfer
                ("ui16", (np.array(10, dtype=np.uint16).astype(">I").tobytes())),
                ("f32", (np.array(100, dtype=np.float32).astype(">f").tobytes())),
                ("f64", (np.array(1000, dtype=np.float64).astype(">d").tobytes())),
                ("s", b"\x00\x00\x00$This is a data test string (pass 0).\x00"),
                ("u", b"\x00\x00\x13http://www.dods.org\x00"),
                ("U", b"\x00\x00\x00\x0ctest unicode"),
            ]
        )

        for key in expected:
            string = expected[key]
            assert xdrdata.startswith(string)
            xdrdata = xdrdata[len(string) :]
        assert xdrdata == b""


class TestDODSResponseSequence(unittest.TestCase):
    """Test the DODS response with a sequence that has a string."""

    def setUp(self):
        """Create a simple WSGI app."""
        app = App(BaseHandler(SimpleSequence))
        self.res = app.get("/.dods?cast.id,cast.lon")

    def test_body(self):
        """Test response body."""
        dds, xdrdata = self.res.body.split(b"\nData:\n", 1)
        dds = dds.decode("ascii")
        self.assertEqual(
            dds,
            """Dataset {
    Sequence {
        String id;
        Int32 lon;
    } cast;
} SimpleSequence;""",
        )

        self.assertEqual(
            xdrdata,
            START_OF_SEQUENCE + b"\x00\x00\x00\x01"  # length of the string (=1)
            b"1\x00\x00\x00"  # string zero-paddded to 4 bytes
            b"\x00\x00\x00d" + START_OF_SEQUENCE + b"\x00\x00\x00\x01"  # lon
            b"2\x00\x00\x00"
            b"\x00\x00\x00\xc8" + END_OF_SEQUENCE,
        )


class TestDODSResponseArray(unittest.TestCase):
    """Test the DODS response with arrays of bytes and strings."""

    def setUp(self):
        """Create a simple WSGI app."""
        self.app = App(BaseHandler(SimpleArray))

    def test_body(self):
        """Test response body."""
        res = self.app.get("/.dods")
        dds, xdrdata = res.body.split(b"\nData:\n", 1)
        dds = dds.decode("ascii")
        self.assertEqual(
            dds,
            """Dataset {
    Byte byte[byte = 5];
    String string[string = 2];
    Int16 short;
} SimpleArray;""",
        )

        self.assertEqual(
            xdrdata,
            b"\x00\x00\x00\x05"  # length of the byte array
            b"\x00\x00\x00\x05"  # length, again
            b"\x00"
            b"\x01"
            b"\x02"
            b"\x03"
            b"\x04"
            b"\x00\x00\x00"  # zero-padding to 4 bytes
            b"\x00\x00\x00\x02"  # length of the array, only once (string)
            b"\x00\x00\x00\x03"  # length of the first string
            b"one\x00"  # string padded to 4 bytes
            b"\x00\x00\x00\x03"
            b"two\x00"
            b"\x00\x00\x00\x01",
        )  # shorts are encoded in 4 bytes

    def test_calculate_size_short(self):
        """Test that size is calculated correctly with shorts."""
        res = self.app.get("/.dods?short")
        self.assertEqual(res.headers["content-length"], "52")


class TestDODSResponseArrayterator(unittest.TestCase):
    """Test the DODS response when encountering an Arrayterator.

    In several cases the DODS response will work on an Arrayterator, which is
    a wrapper over arrays that allows iteration over them in blocks, instead of
    reading everything into memory.

    """

    def test_grid(self):
        """Test a grid."""
        modified = copy.copy(SimpleGrid)
        modified.SimpleGrid.SimpleGrid.data = Arrayterator(
            modified.SimpleGrid.SimpleGrid.data, 2
        )

        app = App(BaseHandler(modified))
        res = app.get("/.dods?SimpleGrid.SimpleGrid")
        header_string = b"""Dataset {
    Structure {
        Int32 SimpleGrid[y = 2][x = 3];
    } SimpleGrid;
} SimpleGrid;
Data:
"""
        self.assertEqual(
            res.body,
            (
                header_string + b"\x00\x00\x00\x06"
                b"\x00\x00\x00\x06"
                b"\x00\x00\x00\x00"
                b"\x00\x00\x00\x01"
                b"\x00\x00\x00\x02"
                b"\x00\x00\x00\x03"
                b"\x00\x00\x00\x04"
                b"\x00\x00\x00\x05"
            ),
        )


class TestDODSResponseNestedSequence(unittest.TestCase):
    """Test the DODS response with nested sequences."""

    def test_iteration(self):
        """Test direct iteration over data."""
        response = DODSResponse(NestedSequence)
        output = b"".join(response)
        self.assertEqual(
            output,
            b"""Dataset {
    Sequence {
        Int32 lat;
        Int32 lon;
        Int32 elev;
        Sequence {
            Int32 time;
            Int32 slp;
            Int32 wind;
        } time_series;
    } location;
} NestedSequence;
Data:
"""
            + START_OF_SEQUENCE
            + b"\x00\x00\x00\x01"
            b"\x00\x00\x00\x01"
            b"\x00\x00\x00\x01" + START_OF_SEQUENCE + b"\x00\x00\x00\n"
            b"\x00\x00\x00\x0b"
            b"\x00\x00\x00\x0c" + START_OF_SEQUENCE + b"\x00\x00\x00\x15"
            b"\x00\x00\x00\x16"
            b"\x00\x00\x00\x17"
            + END_OF_SEQUENCE
            + START_OF_SEQUENCE
            + b"\x00\x00\x00\x02"
            b"\x00\x00\x00\x04"
            b"\x00\x00\x00\x04" + START_OF_SEQUENCE + b"\x00\x00\x00\x0f"
            b"\x00\x00\x00\x10"
            b"\x00\x00\x00\x11"
            + END_OF_SEQUENCE
            + START_OF_SEQUENCE
            + b"\x00\x00\x00\x03"
            b"\x00\x00\x00\x06"
            b"\x00\x00\x00\t" + START_OF_SEQUENCE + b"\x00\x00\x00\x04"
            b"\x00\x00\x00\x08"
            b"\x00\x00\x00\x10" + START_OF_SEQUENCE + b"\x00\x00\x00\x1f"
            b"\x00\x00\x00 "
            b"\x00\x00\x00!" + START_OF_SEQUENCE + b"\x00\x00\x00)"
            b"\x00\x00\x00*"
            b"\x00\x00\x00+" + START_OF_SEQUENCE + b"\x00\x00\x003"
            b"\x00\x00\x004"
            b"\x00\x00\x005" + START_OF_SEQUENCE + b"\x00\x00\x00="
            b"\x00\x00\x00>"
            b"\x00\x00\x00?" + END_OF_SEQUENCE + END_OF_SEQUENCE,
        )

    def test_body(self):
        """Test response body."""
        app = App(BaseHandler(NestedSequence))
        res = app.get("/.dods")
        dds, xdrdata = res.body.split(b"\nData:\n", 1)
        dds = dds.decode("ascii")
        self.assertEqual(
            dds,
            """Dataset {
    Sequence {
        Int32 lat;
        Int32 lon;
        Int32 elev;
        Sequence {
            Int32 time;
            Int32 slp;
            Int32 wind;
        } time_series;
    } location;
} NestedSequence;""",
        )

        self.assertEqual(
            xdrdata,
            START_OF_SEQUENCE + b"\x00\x00\x00\x01"
            b"\x00\x00\x00\x01"
            b"\x00\x00\x00\x01" + START_OF_SEQUENCE + b"\x00\x00\x00\n"
            b"\x00\x00\x00\x0b"
            b"\x00\x00\x00\x0c" + START_OF_SEQUENCE + b"\x00\x00\x00\x15"
            b"\x00\x00\x00\x16"
            b"\x00\x00\x00\x17"
            + END_OF_SEQUENCE
            + START_OF_SEQUENCE
            + b"\x00\x00\x00\x02"
            b"\x00\x00\x00\x04"
            b"\x00\x00\x00\x04" + START_OF_SEQUENCE + b"\x00\x00\x00\x0f"
            b"\x00\x00\x00\x10"
            b"\x00\x00\x00\x11"
            + END_OF_SEQUENCE
            + START_OF_SEQUENCE
            + b"\x00\x00\x00\x03"
            b"\x00\x00\x00\x06"
            b"\x00\x00\x00\t" + START_OF_SEQUENCE + b"\x00\x00\x00\x04"
            b"\x00\x00\x00\x08"
            b"\x00\x00\x00\x10" + START_OF_SEQUENCE + b"\x00\x00\x00\x1f"
            b"\x00\x00\x00 "
            b"\x00\x00\x00!" + START_OF_SEQUENCE + b"\x00\x00\x00)"
            b"\x00\x00\x00*"
            b"\x00\x00\x00+" + START_OF_SEQUENCE + b"\x00\x00\x003"
            b"\x00\x00\x004"
            b"\x00\x00\x005" + START_OF_SEQUENCE + b"\x00\x00\x00="
            b"\x00\x00\x00>"
            b"\x00\x00\x00?" + END_OF_SEQUENCE + END_OF_SEQUENCE,
        )



import multiprocessing
import sys
import time
import warnings

import requests
from werkzeug.serving import run_simple

from ..handlers.lib import BaseHandler
from ..wsgi.ssf import ServerSideFunctions
from .devel import DefaultDataset, LocalTestServer


def run_simple_server(
    application=BaseHandler(DefaultDataset), port=8000, ssl_context=None
):
    application = ServerSideFunctions(application)

    def app_check_for_shutdown(environ, start_response):
        if environ["PATH_INFO"].endswith("shutdown"):
            shutdown_server(environ)
            return shutdown_application(environ, start_response)
        else:
            return application(environ, start_response)

    run_simple("0.0.0.0", port, app_check_for_shutdown, ssl_context=ssl_context)


def shutdown_server(environ):
    pass
    # if 'werkzeug.server.shutdown' not in environ:
    #    raise RuntimeError('Not running the development server')
    # environ['werkzeug.server.shutdown']()


def shutdown_application(environ, start_response):
    start_response("200 OK", [("Content-Type", "text/plain")])
    return [b"Server is shutting down."]


class LocalTestServerSSL(LocalTestServer):
    """
    Simple server instance that can be used to test pydap.
    Relies on multiprocessing and is usually slow (it has to
    start and shutdown which typically takes ~2 sec).

    Usage:
    >>> import numpy as np
    >>> from pydap.handlers.lib import BaseHandler
    >>> from pydap.model import DatasetType, BaseType
    >>> DefaultDataset = DatasetType("Default")
    >>> DefaultDataset["byte"] = BaseType("byte", np.arange(5, dtype="B"))
    >>> DefaultDataset["string"] = BaseType("string", np.array(["one", "two"]))
    >>> DefaultDataset["short"] = BaseType("short", np.array(1, dtype="h"))
    >>> DefaultDataset
    <DatasetType with children 'byte', 'string', 'short'>

    As an instance:
    >>> from pydap.client import open_url
    >>> application = BaseHandler(DefaultDataset)
    >>> with LocalTestServerSSL(application) as server:
    ...     dataset = open_url("http://localhost:%s" % server.port)


    Or by managing connection and deconnection:
    >>> server = LocalTestServerSSL(application)
    >>> server.start()
    >>> dataset = open_url("http://localhost:%s" % server.port)
    >>> dataset
    <DatasetType with children 'byte', 'string', 'short'>
    >>> print(dataset['byte'].data[:])
    [0 1 2 3 4]
    >>> print(dataset['string'].data[:])
    [b'one' b'two']
    >>> print(dataset['short'].data[:])
    1
    >>> server.shutdown()
    """

    def __init__(
        self,
        application=BaseHandler(DefaultDataset),
        port=None,
        wait=0.5,
        polling=1e-2,
        as_process=False,
        ssl_context=None,
    ):
        super(LocalTestServerSSL, self).__init__(
            application, port, wait, polling, as_process
        )
        self._ssl_context = ssl_context

    @property
    def url(self):
        if self._ssl_context is None:
            return "http://{0}:{1}/".format(self._address, self.port)
        else:
            return "https://{0}:{1}/".format(self._address, self.port)

    def start(self):
        if sys.platform in ["darwin", "win32"]:
            # see https://github.com/python/cpython/issues/77906
            # no long term solution, simply temporaty fix
            ctx = multiprocessing.get_context("fork")
            Process = ctx.Process
        else:
            Process = multiprocessing.Process

        # Start a simple WSGI server:
        self._server = Process(
            target=run_simple_server,
            args=(self.application, self.port, self._ssl_context),
        )
        self._server.start()
        # Wait a little while for the server to start:
        self.poll_server()

    # https://werkzeug.palletsprojects.com/en/2.2.x/serving/#shutting-down-the-server
    def shutdown(self):
        # Shutdown the server:
        url = "http://0.0.0.0:%s/shutdown"
        if self._ssl_context is not None:
            url = url.replace("http", "https")
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            requests.head(url % self.port, verify=False)
        time.sleep(self._wait)
        # self._server.join()
        self._server.terminate()
        del self._server



import unittest

from webtest import TestApp as App


def MyApp(environ, start_response):
    start_response("200 OK", [("content-type", "text/plain; charset=utf-8")])
    yield "foo\n".encode("utf-8")
    yield "bar\n".encode("utf-8")
    yield "baz\n".encode("utf-8")


class MyTest(unittest.TestCase):
    def setUp(self):
        self.app = App(MyApp)

    def test_webtest(self):
        expected = "foo\nbar\nbaz\n"
        self.assertEqual(self.app.get("/").text, expected)



"""Test the HTML response from pydap."""

import unittest
from collections import OrderedDict

from bs4 import BeautifulSoup
from jinja2 import DictLoader, Environment
from webob import Request
from webob.headers import ResponseHeaders
from webtest import TestApp as App

from pydap.handlers.lib import BaseHandler
from pydap.lib import __version__, walk
from pydap.tests.datasets import SimpleGrid, SimpleGroup, VerySimpleSequence


class TestHTMLResponseSequence(unittest.TestCase):
    """Test the HTML response with a dataset containing a sequence."""

    def setUp(self):
        """Create a simple app and request HTML response."""
        self.app = App(BaseHandler(VerySimpleSequence))

    def test_status(self):
        """Test the status code."""
        res = self.app.get("/.html")
        self.assertEqual(res.status, "200 OK")

    def test_content_type(self):
        """Test the content type header."""
        res = self.app.get("/.html")
        self.assertEqual(res.content_type, "text/html")

    def test_charset(self):
        """Test the charset."""
        res = self.app.get("/.html")
        self.assertEqual(res.charset, "utf-8")

    def test_headers(self):
        """Test the response headers."""
        res = self.app.get("/.html")
        self.assertEqual(
            res.headers,
            ResponseHeaders(
                [
                    ("OPeNDAP-Server", "pydap/" + __version__),
                    ("Content-description", "DAP_form"),
                    ("Content-type", "text/html; charset=utf-8"),
                    ("Content-Length", "5884"),
                ]
            ),
        )

    def test_body(self):
        """Test the HTML response.

        We use BeautifulSoup to parse the response, and check for some
        elements that should be there.

        """
        res = self.app.get("/.html")
        soup = BeautifulSoup(res.text, "html.parser")

        self.assertEqual(soup.title.string, "Dataset http://localhost/.html")
        self.assertEqual(soup.form["action"], "http://localhost/.html")
        self.assertEqual(soup.form["method"], "POST")

        # check that all variables are present
        ids = [var.id for var in walk(VerySimpleSequence)]
        for h2, id_ in zip(soup.find_all("h2"), ids):
            self.assertEqual(h2.string, id_)

    def test_post(self):
        """Test that the data redirect works."""
        res = self.app.post("/.html")
        self.assertEqual(res.status, "303 See Other")
        self.assertEqual(res.location, "http://localhost/.ascii")

    def test_post_selection(self):
        """Test that the data redirect works with a subset request."""
        res = self.app.post("/.html", {"sequence.byte": "on"})
        self.assertEqual(res.location, "http://localhost/.ascii?sequence.byte")

    def test_post_multiple_selection(self):
        """Test that the data redirect works with a subset request."""
        # from nose.tools import set_trace; set_trace()
        res = self.app.post(
            "/.html", OrderedDict([("sequence.byte", "on"), ("sequence.int", "on")])
        )
        self.assertEqual(
            res.location, "http://localhost/.ascii?sequence.byte,sequence.int"
        )

    def test_filter(self):
        """Test that filtering the data works."""
        res = self.app.post(
            "/.html",
            {
                "var1_sequence": "sequence.byte",
                "op_sequence": ">",
                "var2_sequence": "10",
            },
        )
        self.assertEqual(res.location, "http://localhost/.ascii?&sequence.byte>10")


class TestHTMLResponseGrid(unittest.TestCase):
    """Test the HTML response with a dataset containing a sequence."""

    def setUp(self):
        """Create a simple app and request HTML response."""
        self.app = App(BaseHandler(SimpleGrid))

    def test_filter(self):
        """Test filtering the grid."""
        res = self.app.post(
            "/.html",
            {
                "SimpleGrid": "on",
                "SimpleGrid[0]": "0",
                "SimpleGrid[1]": "0",
            },
        )
        self.assertEqual(res.location, "http://localhost/.ascii?SimpleGrid[0][0]")


class TestHTMLTemplate(unittest.TestCase):
    """Test that global environment is used if available.

    The HTML response has its own environment and loader, for when used in
    standalone mode. If used with a pydap server that defines its own
    environment, the HTML response will reuse the global environment after
    injecting its loader as a backup. This allows the HTML response to work in
    case the global environment does not have the HTML templates.

    """

    def test_environ_loader_with_template(self):
        """Test that global environment is used."""
        loader = DictLoader({"html.html": "global"})
        env = Environment(loader=loader)

        app = BaseHandler(VerySimpleSequence)
        req = Request.blank("/.html")
        req.environ["pydap.jinja2.environment"] = env
        res = req.get_response(app)
        self.assertEqual(res.text, "global")

    def test_environ_loader_without_template(self):
        """Test that global environment is used."""
        env = Environment()
        self.assertIsNone(env.loader)

        app = BaseHandler(VerySimpleSequence)
        req = Request.blank("/.html")
        req.environ["pydap.jinja2.environment"] = env
        res = req.get_response(app)
        self.assertNotEqual(res.text, "global")


class TestHTMLResponseSimpleGroup(unittest.TestCase):
    """Test the HTML response with a dataset containing a sequence."""

    def setUp(self):
        """Create a simple app and request HTML response."""
        self.app = App(BaseHandler(SimpleGroup))

    def test_status(self):
        """Test the status code."""
        res = self.app.get("/.html")
        self.assertEqual(res.status, "200 OK")

    def test_headers(self):
        """Test the response headers."""
        res = self.app.get("/.html")
        self.assertEqual(
            res.headers,
            ResponseHeaders(
                [
                    ("OPeNDAP-Server", "pydap/" + __version__),
                    ("Content-description", "DAP_form"),
                    ("Content-type", "text/html; charset=utf-8"),
                    ("Content-Length", "8101"),
                ]
            ),
        )



"""Test the DAS response."""

import unittest

from webob.headers import ResponseHeaders
from webtest import TestApp as App

from pydap.handlers.lib import BaseHandler
from pydap.lib import __version__
from pydap.responses.das import das
from pydap.tests.datasets import FaultyGrid, SimpleGrid, SimpleSequence, SimpleStructure


class TestDASResponseSequence(unittest.TestCase):
    """Test the DAS response from a sequence."""

    def setUp(self):
        """Create a simple WSGI app."""
        app = App(BaseHandler(SimpleSequence))
        self.res = app.get("/.das")

    def test_dispatcher(self):
        """Test the single dispatcher."""
        with self.assertRaises(StopIteration):
            das(None)

    def test_status(self):
        """Test the status code."""
        self.assertEqual(self.res.status, "200 OK")

    def test_content_type(self):
        """Test the content type."""
        self.assertEqual(self.res.content_type, "text/plain")

    def test_charset(self):
        """Test the charset."""
        self.assertEqual(self.res.charset, "ascii")

    def test_headers(self):
        """Test the response headers."""
        self.assertEqual(
            self.res.headers,
            ResponseHeaders(
                [
                    ("OPeNDAP-Server", "pydap/" + __version__),
                    ("Content-description", "dods_das"),
                    ("Content-type", "text/plain; charset=ascii"),
                    ("Access-Control-Allow-Origin", "*"),
                    (
                        "Access-Control-Allow-Headers",
                        "Origin, X-Requested-With, Content-Type",
                    ),
                    ("Content-Length", "510"),
                ]
            ),
        )

    def test_body(self):
        """Test the generated DAS response."""
        self.assertEqual(
            self.res.text,
            """Attributes {
    String description "A simple sequence for testing.";
    nested {
        Int32 value 42;
    }
    cast {
        id {
        }
        lon {
            String axis "X";
        }
        lat {
            String axis "Y";
        }
        depth {
            String axis "Z";
        }
        time {
            String axis "T";
            String units "days since 1970-01-01";
        }
        temperature {
        }
        salinity {
        }
        pressure {
        }
    }
}
""",
        )


class TestDASResponseGrid(unittest.TestCase):
    """Test the DAS response from a grid."""

    def test_body(self):
        """Test the generated DAS response."""
        app = App(BaseHandler(SimpleGrid))
        res = app.get("/.das")
        self.assertEqual(
            res.text,
            """Attributes {
    String description "A simple grid for testing.";
    SimpleGrid {
    }
    x {
        String axis "X";
        String units "degrees_east";
    }
    y {
        String axis "Y";
        String units "degrees_north";
    }
}
""",
        )


class TestDASResponseFaultyGrid(unittest.TestCase):
    """Test the DAS response from a grid with empty properties."""

    def test_body(self):
        """Test the generated DAS response."""
        app = App(BaseHandler(FaultyGrid))
        res = app.get("/.das")
        self.assertEqual(
            res.text,
            """Attributes {
    String description "A faulty grid for testing.";
    FaultyGrid {
    }
    x {
        String axis "X";
        Int32 code 1;
    }
    y {
        String axis "Y";
    }
}
""",
        )


class TestDASResponseStructure(unittest.TestCase):
    """The the DAS response from a structure."""

    def test_body(self):
        """Test the generated DAS response."""
        app = App(BaseHandler(SimpleStructure))
        res = app.get("/.das")
        self.assertEqual(
            res.text,
            """Attributes {
    types {
        String key "value";
        nested {
            String string "bar";
            Int32 list 42, 43;
            Int32 array 1;
            Float64 float 1000;
        }
        b {
        }
        ub {
        }
        i32 {
        }
        ui32 {
        }
        i16 {
        }
        ui16 {
        }
        f32 {
        }
        f64 {
        }
        s {
        }
        u {
        }
        U {
        }
    }
}
""",
        )



"""Test the DDS response."""

import unittest

from webob.headers import ResponseHeaders
from webtest import TestApp as App

from pydap.handlers.lib import BaseHandler
from pydap.lib import __version__
from pydap.responses.dds import dds
from pydap.tests.datasets import SimpleGrid, SimpleSequence, SimpleStructure


class TestDDSResponseSequence(unittest.TestCase):
    """Test DDS response from sequences."""

    def setUp(self):
        """Create a simple WSGI app."""
        app = App(BaseHandler(SimpleSequence))
        self.res = app.get("/.dds")

    def test_dispatcher(self):
        """Test the single dispatcher."""
        with self.assertRaises(StopIteration):
            dds(None)

    def test_status(self):
        """Test the status code."""
        self.assertEqual(self.res.status, "200 OK")

    def test_content_type(self):
        """Test the content type."""
        self.assertEqual(self.res.content_type, "text/plain")

    def test_charset(self):
        """Test the charset."""
        self.assertEqual(self.res.charset, "ascii")

    def test_headers(self):
        """Test the headers from the response."""
        self.assertEqual(
            self.res.headers,
            ResponseHeaders(
                [
                    ("OPeNDAP-Server", "pydap/" + __version__),
                    ("Content-description", "dods_dds"),
                    ("Content-type", "text/plain; charset=ascii"),
                    ("Access-Control-Allow-Origin", "*"),
                    (
                        "Access-Control-Allow-Headers",
                        "Origin, X-Requested-With, Content-Type",
                    ),
                    ("Content-Length", "228"),
                ]
            ),
        )

    def test_body(self):
        """Test the generated DDS response."""
        self.assertEqual(
            self.res.text,
            """Dataset {
    Sequence {
        String id;
        Int32 lon;
        Int32 lat;
        Int32 depth;
        Int32 time;
        Int32 temperature;
        Int32 salinity;
        Int32 pressure;
    } cast;
} SimpleSequence;
""",
        )


class TestDDSResponseGrid(unittest.TestCase):
    """Test DDS response from grids."""

    def test_body(self):
        """Test the generated DDS response."""
        app = App(BaseHandler(SimpleGrid))
        res = app.get("/.dds")
        self.assertEqual(
            res.text,
            """Dataset {
    Grid {
        Array:
            Int32 SimpleGrid[y = 2][x = 3];
        Maps:
            Int32 x[x = 3];
            Int32 y[y = 2];
    } SimpleGrid;
    Int32 x[x = 3];
    Int32 y[y = 2];
} SimpleGrid;
""",
        )


class TestDDSResponseStructure(unittest.TestCase):
    """Test DDS response from structures."""

    def test_body(self):
        """Test the generated DDS response."""
        app = App(BaseHandler(SimpleStructure))
        res = app.get("/.dds")
        self.assertEqual(
            res.text,
            """Dataset {
    Structure {
        Int16 b;
        Byte ub;
        Int32 i32;
        UInt32 ui32;
        Int16 i16;
        UInt16 ui16;
        Float32 f32;
        Float64 f64;
        String s;
        String u;
        String U;
    } types;
} SimpleStructure;
""",
        )



# Define which file to ignore in tests:
collect_ignore = ["setup.py", "bootstrap.py", "docs/conf.py"]

# This line should be deleted once the online documentation is
# fixed:
collect_ignore.append("docs/client.rst")

# These lines should be deleted when all examples use local files:
collect_ignore.append("docs/index.rst")
collect_ignore.append("src/pydap/client.py")



"""
This test module uses a simple development server
to test http-specific options like timeout
or caching.

It is based on simple data but with more handlers
it could work with more data formats.
"""

import numpy as np
import pytest

from pydap.client import open_url
from pydap.handlers.lib import BaseHandler
from pydap.model import BaseType, DatasetType, SequenceType
from pydap.server.devel import LocalTestServer

server = pytest.mark.server


@pytest.fixture
def sequence_type_data():
    """
    Simple sequence test data
    """
    data = [
        (10, 15.2, "Diamond_St"),
        (11, 13.1, "Blacktail_Loop"),
        (12, 13.3, "Platinum_St"),
        (13, 12.1, "Kodiak_Trail"),
    ]
    dtype = [("index", "<i4"), ("temperature", "<f8"), ("station", "S40")]
    seq = SequenceType("sequence")
    seq["index"] = BaseType("index")
    seq["temperature"] = BaseType("temperature")
    seq["station"] = BaseType("station")
    seq.data = np.array(data, dtype=dtype)
    return seq


@server
def test_open(sequence_type_data):
    """Test that LocalTestServer works properly"""
    TestDataset = DatasetType("Test")
    TestDataset["sequence"] = sequence_type_data
    with LocalTestServer(BaseHandler(TestDataset)) as server:
        dataset = open_url(server.url)
        seq = dataset["sequence"]
        retrieved_data = [line for line in seq]

    np.testing.assert_array_equal(
        np.array(retrieved_data, dtype=sequence_type_data.data.dtype),
        np.array(sequence_type_data.data[:], dtype=sequence_type_data.data.dtype),
    )


@server
def test_netcdf(sequence_type_data):
    """
    Test that LocalTestServer works properly and that it works well with
    netcdf4-python.
    """
    TestDataset = DatasetType("Test")
    TestDataset["float"] = BaseType("float", np.array(1, dtype=np.float32))
    with TestDataset.to_netcdf() as ds:
        assert "float" in ds.variables
        assert ds["float"].dtype == np.float32
        assert ds["float"][:] == np.array(1, dtype=np.float32)


# @server
# def test_timeout(sequence_type_data):
#     """Test that timeout works properly"""
#     TestDataset = DatasetType('Test')
#     TestDataset['sequence'] = sequence_type_data
#     TestDataset['byte'] = BaseType('byte', 0)
#     application = BaseHandler(TestDataset)

#     # Explictly add latency on the devel server
#     # to guarantee that it timeouts
#     def wrap_mocker(func):
#         def mock_add_latency(*args, **kwargs):
#             time.sleep(1e-1)
#             return func(*args, **kwargs)
#         return mock_add_latency

#     application = wrap_mocker(application)
#     with LocalTestServer(application) as server:
#         url = ("http://0.0.0.0:%s/" % server.port)

#         # test open_url
#         assert open_url(url) == TestDataset
#         with pytest.raises(HTTPError) as e:
#             open_url(url, timeout=1e-5)
#         assert 'Timeout' in str(e)

#         # test open_dods
#         with pytest.raises(HTTPError):
#             open_dods(url + '.dods?sequence', timeout=1e-5)
#         assert 'Timeout' in str(e)

#         # test sequenceproxy
#         dataset = open_url(url)
#         seq = dataset['sequence']
#         assert isinstance(seq.data, SequenceProxy)
#         # Change the timeout of the sequence proxy:
#         seq.data.timeout = 1e-5
#         with pytest.raises(HTTPError) as e:
#             next(seq.iterdata())
#         assert 'Timeout' in str(e)

#         # test baseproxy:
#         dat = dataset['byte']
#         assert isinstance(dat.data, BaseProxy)
#         # Change the timeout of the baseprox proxy:
#         dat.data.timeout = 1e-5
#         with pytest.raises(HTTPError) as e:
#             dat[:]
#         assert 'Timeout' in str(e)



"""Test the pydap client."""

import os

import numpy as np
import pytest

from pydap.client import open_dods_url, open_file, open_url
from pydap.handlers.lib import BaseHandler
from pydap.tests.datasets import SimpleGrid, SimpleSequence, SimpleStructure

DODS = os.path.join(os.path.dirname(__file__), "data/test.01.dods")
DAS = os.path.join(os.path.dirname(__file__), "data/test.01.das")


@pytest.fixture
def sequence_app():
    return BaseHandler(SimpleSequence)


@pytest.fixture
def structure_app():
    return BaseHandler(SimpleStructure)


@pytest.mark.client
def test_open_url(sequence_app):
    """Open an URL and check dataset keys."""
    dataset = open_url("http://localhost:8001/", sequence_app)
    assert list(dataset.keys()) == ["cast"]


@pytest.fixture
def remote_url():
    return "http://test.opendap.org/opendap/hyrax/data/"


def test_open_url_dap4(remote_url):
    base_url = remote_url + "nc/test.nc"
    data_original = open_url(base_url)

    # test single data point
    constrain1 = "dap4.ce=/s33[0][0]"
    data_dap4 = open_url(base_url + "?" + constrain1)
    assert data_dap4["s33"][:].data == data_original["s33"][0, 0].data

    # subset of vars by indexes
    var1 = "/s33[0:1:2][0:1:2];"
    var2 = "/br34[0:1:1][0:1:2][0:1:3];"
    var3 = "/s113[0:1:0][0:1:0][0:1:2]"
    Vars = [var1, var2, var3]

    url = base_url + "?dap4.ce=" + var1 + var2 + var3
    dataset = open_url(url)
    # check [vars1, vars2, vars3] only in dataset
    assert len(dataset.keys()) == len(Vars)


def test_open_url_dap4_shape():
    url = "http://test.opendap.org:8080/opendap/"
    filename = "netcdf/examples/200803061600_HFRadar_USEGC_6km_rtv_SIO.nc"
    CE = "?dap4.ce=/lon[100:1:199]"
    ds_ce = open_url(url + filename + CE)
    data = ds_ce["lon"][:]
    assert data.shape == (100,)


def test_open_url_seqCE(remote_url):
    seq_url = remote_url + "ff/gsodock.dat"
    data_original = open_url(seq_url)
    # get name of Sequence
    seq_name = [key for key in data_original.keys()][0]

    # add constraint expression
    Vars = ["Time", "Sea_Temp"]
    Value = 35234.1
    projection = "URI_GSO-Dock.Time,URI_GSO-Dock.Sea_Temp"
    selection = "URI_GSO-Dock.Time<" + str(Value)
    seq_url_CE = seq_url + "?" + projection + "&" + selection

    dsCE = open_url(seq_url_CE)

    # assert projection works within sequence
    assert set([var for var in dsCE[seq_name].keys()]) == set(Vars)

    # assert selection works within sequence

    assert max([var for var in dsCE[seq_name]["Time"]]) < Value
    assert len([var for var in dsCE[seq_name]["Time"]]) < len(
        [var for var in data_original[seq_name]["Time"]]
    )


@pytest.mark.client
def test_open_dods():
    """Open a file downloaded from the test server with the DAS."""
    dataset = open_file(DODS)

    # test data
    assert dataset.data == [
        0,
        1,
        0,
        0,
        0,
        0.0,
        1000.0,
        "This is a data test string (pass 0).",
        "http://www.dods.org",
    ]

    # test attributes
    assert dataset.attributes == {}
    assert dataset.i32.attributes == {}
    assert dataset.b.attributes == {}


@pytest.mark.client
def test_open_dods_das():
    """Open a file downloaded from the test server with the DAS."""
    dataset = open_file(DODS, DAS)

    # test data
    assert dataset.data == [
        0,
        1,
        0,
        0,
        0,
        0.0,
        1000.0,
        "This is a data test string (pass 0).",
        "http://www.dods.org",
    ]

    # test attributes
    assert dataset.i32.units == "unknown"
    assert dataset.i32.Description == "A 32 bit test server int"
    assert dataset.b.units == "unknown"
    assert dataset.b.Description == "A test byte"
    assert dataset.Facility["DataCenter"] == "COAS Environmental Computer Facility"
    assert dataset.Facility["PrincipleInvestigator"] == ["Mark Abbott", "Ph.D"]
    assert dataset.Facility["DrifterType"] == "MetOcean WOCE/OCM"


@pytest.mark.client
def test_open_dods_16bits(sequence_app):
    """Open the dods response from a server.

    Note that here we cannot simply compare ``dataset.data`` with the
    original data ``SimpleSequence.data``, because the original data
    contains int16 values which are transmitted as int32 in the DAP spec.

    """
    dataset = open_dods_url(".dods", application=sequence_app)
    assert list(dataset.data) == [
        [
            ("1", 100, -10, 0, -1, 21, 35, 0),
            ("2", 200, 10, 500, 1, 15, 35, 100),
        ]
    ]

    # attributes should be empty
    assert dataset.attributes == {}


@pytest.mark.client
def test_open_dods_with_attributes(sequence_app):
    """Open the dods response together with the das response."""
    dataset = open_dods_url(".dods", metadata=True, application=sequence_app)
    assert dataset.NC_GLOBAL == {}
    assert dataset.DODS_EXTRA == {}
    assert dataset.description == "A simple sequence for testing."
    assert dataset.cast.lon.axis == "X"
    assert dataset.cast.lat.axis == "Y"
    assert dataset.cast.depth.axis == "Z"
    assert dataset.cast.time.axis == "T"
    assert dataset.cast.time.units == "days since 1970-01-01"


@pytest.fixture(scope="session")
def ssf_app():
    """Test the local implementation of server-side functions.

    Calling server-side functions is implemented using a lazy mechanism where
    arbitrary names are mapped to remove calls. The resulting dataset is only
    evaluated when ``__getitem__`` or ``__getattr__`` are called, allowing
    nested calls to be evaluated only once.

    """
    from pydap.wsgi.ssf import ServerSideFunctions

    return ServerSideFunctions(BaseHandler(SimpleGrid))


def test_original(ssf_app):
    """Test an unmodified call, without function calls."""
    original = open_url("/", application=ssf_app)
    assert original.SimpleGrid.SimpleGrid.shape == (2, 3)


def test_first_axis(ssf_app):
    """Test mean over the first axis."""
    original = open_url("/", application=ssf_app)
    dataset = original.functions.mean(original.SimpleGrid, 0)
    assert dataset.SimpleGrid.SimpleGrid.shape == (3,)
    np.testing.assert_array_equal(
        dataset.SimpleGrid.SimpleGrid.data, np.array([1.5, 2.5, 3.5])
    )


def test_second_axis(ssf_app):
    """Test mean over the second axis."""
    original = open_url("/", application=ssf_app)
    dataset = original.functions.mean(original.SimpleGrid, 1)
    assert dataset.SimpleGrid.SimpleGrid.shape == (2,)
    np.testing.assert_array_equal(
        dataset.SimpleGrid.SimpleGrid.data, np.array([1.0, 4.0])
    )


def test_lazy_evaluation_getitem(ssf_app):
    """Test that the dataset is only loaded when accessed."""
    original = open_url("/", application=ssf_app)
    dataset = original.functions.mean(original.SimpleGrid, 0)
    assert dataset.dataset is None
    dataset["SimpleGrid"]
    assert dataset.dataset is not None


def test_lazy_evaluation_getattr(ssf_app):
    """Test that the dataset is only loaded when accessed."""
    original = open_url("/", application=ssf_app)
    dataset = original.functions.mean(original.SimpleGrid, 0)
    assert dataset.dataset is None
    dataset.SimpleGrid
    assert dataset.dataset is not None


def test_nested_call(ssf_app):
    """Test nested calls."""
    original = open_url("/", application=ssf_app)
    dataset = original.functions.mean(
        original.functions.mean(original.SimpleGrid, 0), 0
    )
    assert dataset["SimpleGrid"]["SimpleGrid"].shape == ()
    np.testing.assert_array_equal(dataset.SimpleGrid.SimpleGrid.data, np.array(2.5))


def test_axis_mean(ssf_app):
    """Test the mean over an axis, returning a scalar."""
    original = open_url("/", application=ssf_app)
    dataset = original.functions.mean(original.SimpleGrid.x)
    assert dataset.x.shape == ()
    np.testing.assert_array_equal(dataset.x.data, np.array(1.0))


@pytest.mark.client
def test_int16(structure_app):
    """Test that 16-bit values are represented correctly.

    Even though the DAP transfers int16 and uint16 as 32 bits, we want them to
    be represented locally using the correct type.

    Load an int16 -> should yield '>i2' type."""
    dataset = open_url("http://localhost:8001/", structure_app)
    assert dataset.types.i16.dtype == np.dtype(">i2")


@pytest.mark.client
def test_uint16(structure_app):
    """Test that 16-bit values are represented correctly.

    Even though the DAP transfers int16 and uint16 as 32 bits, we want them to
    be represented locally using the correct type.

    Load an uint16 -> should yield '>u2' type."""
    dataset = open_url("http://localhost:8001/", structure_app)
    assert dataset.types.ui16.dtype == np.dtype(">u2")



"""Test pydap base exception."""

import unittest

from pydap.exceptions import DapError


class TestExceptions(unittest.TestCase):
    """Test pydap base exception.

    Other exceptions behave exactly like the superclass, differing only by
    name, so no testing is required.

    """

    def test_dap_error(self):
        """Test the base exception."""
        exc = DapError("This is a test.")

        self.assertEqual(exc.value, "This is a test.")
        self.assertEqual(str(exc), repr("This is a test."))



import os

import numpy as np
import pytest
import requests

import pydap.net
from pydap.cas import esgf
from pydap.client import open_url

url = (
    "http://esgf-data.ucar.edu/thredds/dodsC/cmip5/output1/"
    "NCAR/CCSM4/historical/fx/atmos/fx/r0i0p0/v20130312/orog/"
    "orog_fx_CCSM4_historical_r0i0p0.nc"
)
test_url = url + ".dods?orog[0:1:4][0:1:4]"


@pytest.mark.auth
@pytest.mark.prod_url
@pytest.mark.skipif(
    not (
        os.environ.get("OPENID_ESGF_NO_REG") and os.environ.get("PASSWORD_ESGF_NO_REG")
    ),
    reason=("Without auth credentials, " "this test cannot work"),
)
def test_registration_esgf_auth():
    """
    Attempt to access a ESGF OPENDAP link for which
    the user has not yet selected a registration group.
    This procedure has to be completed only once per project
    over the lifetime of an ESGF OPENID

    Requires OPENID_ESGF_NO_REG and PASSWORD_ESGF_NO_REG
    environment variables. These must be associated with credentials
    where no group was selected.
    """
    with pytest.raises(UserWarning):
        esgf.setup_session(
            os.environ.get("OPENID_ESGF_NO_REG"),
            os.environ.get("PASSWORD_ESGF_NO_REG"),
            check_url=url,
        )


@pytest.mark.auth
@pytest.mark.prod_url
@pytest.mark.skipif(
    not (os.environ.get("OPENID_ESGF") and os.environ.get("PASSWORD_ESGF")),
    reason=("Without auth credentials, " "this test cannot work"),
)
def test_basic_esgf_auth():
    """
    Set up PyDAP to use the ESGF request() function.

    The intent here is to ensure that pydap.net is able to
    open and url if and only if requests is able to
    open the same url.
    """
    session = esgf.setup_session(
        os.environ.get("OPENID_ESGF"), os.environ.get("PASSWORD_ESGF"), check_url=url
    )

    res = requests.get(test_url, cookies=session.cookies)
    assert res.status_code == 200
    res.close()

    res = pydap.net.follow_redirect(test_url, session=session)
    assert res.status_code == 200


@pytest.mark.auth
@pytest.mark.prod_url
@pytest.mark.skipif(
    not (os.environ.get("OPENID_ESGF") and os.environ.get("PASSWORD_ESGF")),
    reason=("Without auth credentials, " "this test cannot work"),
)
def test_dimension_esgf_query():
    session = esgf.setup_session(
        os.environ.get("OPENID_ESGF"), os.environ.get("PASSWORD_ESGF"), check_url=url
    )

    # Ensure authentication:
    res = pydap.net.follow_redirect(test_url, session=session)
    assert res.status_code == 200

    dataset = open_url(url, session=session)
    data = dataset["lon"][:5]
    expected_data = np.array([0.0, 1.25, 2.5, 3.75, 5.0])
    assert np.isclose(data, expected_data).all()


@pytest.mark.auth
@pytest.mark.prod_url
@pytest.mark.skipif(
    not (os.environ.get("OPENID_ESGF") and os.environ.get("PASSWORD_ESGF")),
    reason=("Without auth credentials, " "this test cannot work"),
)
def test_variable_esgf_query():
    session = esgf.setup_session(
        os.environ.get("OPENID_ESGF"), os.environ.get("PASSWORD_ESGF"), check_url=url
    )
    # Ensure authentication:
    res = pydap.net.follow_redirect(test_url, session=session)
    assert res.status_code == 200

    dataset = open_url(url, session=session, output_grid=False)
    data = dataset["orog"][103:105, 100:102]
    expected_data = [[271.36645508, 166.85339355], [304.22286987, 178.85267639]]
    assert np.isclose(data, expected_data).all()


@pytest.mark.auth
@pytest.mark.prod_url
@pytest.mark.skipif(
    not (
        os.environ.get("OPENID_ESGF_CEDA")
        and os.environ.get("USERNAME_ESGF_CEDA")
        and os.environ.get("PASSWORD_ESGF_CEDA")
    ),
    reason=("Without auth credentials, " "this test cannot work"),
)
def test_variable_esgf_query_ceda():
    session = esgf.setup_session(
        os.environ.get("OPENID_ESGF_CEDA"),
        os.environ.get("PASSWORD_ESGF_CEDA"),
        check_url=url,
        username=os.environ.get("USERNAME_ESGF_CEDA"),
    )
    # Ensure authentication:
    res = pydap.net.follow_redirect(test_url, session=session)
    assert res.status_code == 200

    dataset = open_url(url, session=session, output_grid=False)
    data = dataset["orog"][103:105, 100:102]
    expected_data = [[271.36645508, 166.85339355], [304.22286987, 178.85267639]]
    assert np.isclose(data, expected_data).all()
